"""Core pydoctor objects.

The two core objects are L{Documentable} and L{System}.  Instances of
(subclasses of) L{Documentable} represent the documentable 'things' in the
system being documented.  An instance of L{System} represents the whole system
being documented -- a System is a bad of Documentables, in some sense.
"""
from __future__ import annotations

import abc
import ast
from itertools import chain
from collections import defaultdict
import datetime
import importlib
import sys
import textwrap
import types
from enum import Enum
from inspect import signature, Signature
from pathlib import Path
from typing import (
    TYPE_CHECKING, Any, Collection, Dict, Iterable, Iterator, List, Mapping, Callable, 
    Optional, Sequence, Set, Tuple, Type, TypeVar, Union, cast, overload
)
from graphlib import TopologicalSorter
from urllib.parse import quote

import attr

from pydoctor.options import Options
from pydoctor import factory, qnmatch, utils, linker, astutils, mro
from pydoctor.epydoc.markup import ParsedDocstring
from pydoctor.sphinx import CacheT, SphinxInventory

if TYPE_CHECKING:
    from typing import Literal, Protocol, TypeAlias
    from pydoctor.astbuilder import ASTBuilder, DocumentableT, SyntaxTreeParser
else:
    Literal = {True: bool, False: bool}
    ASTBuilder = Protocol = object

T = TypeVar('T')

# originally when I started to write pydoctor I had this idea of a big
# tree of Documentables arranged in an almost arbitrary tree.
#
# this was misguided.  the tree structure is important, to be sure,
# but the arrangement of the tree is far from arbitrary and there is
# at least some code that now relies on this.  so here's a list:
#
#   Packages can contain Packages and Modules
#   Modules can contain Functions and Classes
#   Classes can contain Functions (in this case they get called Methods) and
#       Classes
#   Functions can't contain anything.


class LineFromAst(int):
    "Simple L{int} wrapper for linenumbers coming from ast analysis."

class LineFromDocstringField(int):
    "Simple L{int} wrapper for linenumbers coming from docstrings."

class DocLocation(Enum):
    OWN_PAGE = 1
    PARENT_PAGE = 2
    # Nothing uses this yet.  Parameters will one day.
    #UNDER_PARENT_DOCSTRING = 3


class ProcessingState(Enum):
    UNPROCESSED = 0
    PROCESSING = 1
    PROCESSED = 2


class PrivacyClass(Enum):
    """L{Enum} containing values indicating how private an object should be.

    @cvar HIDDEN: Don't show the object at all.
    @cvar PRIVATE: Show, but de-emphasize the object.
    @cvar PUBLIC: Show the object as normal.
    """

    HIDDEN = 0
    PRIVATE = 1
    PUBLIC = 2
    # For compatibility
    VISIBLE = PUBLIC

class DocumentableKind(Enum):
    """
    L{Enum} containing values indicating the possible object types.

    @note: Presentation order is derived from the enum values
    """
    NAMESPACE_PACKAGE   = 1001
    PACKAGE             = 1000
    MODULE              = 900
    CLASS               = 800
    INTERFACE           = 850
    EXCEPTION           = 750
    CLASS_METHOD        = 700
    STATIC_METHOD       = 600
    METHOD              = 500
    FUNCTION            = 400
    CONSTANT            = 310
    TYPE_VARIABLE       = 306
    TYPE_ALIAS          = 305
    CLASS_VARIABLE      = 300
    SCHEMA_FIELD        = 220
    ATTRIBUTE           = 210
    INSTANCE_VARIABLE   = 200
    PROPERTY            = 150
    VARIABLE            = 100

class Documentable:
    """An object that can be documented.

    The interface is a bit ridiculously wide.

    @ivar docstring: The object's docstring.  But also see docsources.
    @ivar system: The system the object is part of.

    """
    docstring: Optional[str] = None
    parsed_docstring: Optional[ParsedDocstring] = None
    parsed_summary: Optional[ParsedDocstring] = None
    parsed_type: Optional[ParsedDocstring] = None
    docstring_lineno = 0
    linenumber: LineFromAst | LineFromDocstringField | Literal[0] = 0
    source_href: Optional[str] = None
    kind: Optional[DocumentableKind] = None

    documentation_location = DocLocation.OWN_PAGE
    """Page location where we are documented."""

    def __init__(
            self, system: 'System', name: str,
            parent: Optional['Documentable'] = None,
            source_path: Optional[Path] = None
            ):
        if source_path is None and parent is not None:
            source_path = parent.source_path
        self.system = system
        self.name = name
        self.parent = parent
        self.parentMod: Optional[Module] = None
        self.source_path: Optional[Path] = source_path
        self.extra_info: List[ParsedDocstring] = []
        """
        A list to store extra informations about this documentable, as L{ParsedDocstring}.
        """
        self.setup()

    @property
    def doctarget(self) -> 'Documentable':
        return self

    def setup(self) -> None:
        self.contents: Dict[str, Documentable] = {}
        self._linker: Optional['linker.DocstringLinker'] = None

    def setDocstring(self, node: astutils.Str) -> None:
        lineno, doc = astutils.extract_docstring(node)
        self._setDocstringValue(doc, lineno)

    def _setDocstringValue(self, doc:str, lineno:int) -> None:
        if self.docstring or self.parsed_docstring: # some object have a parsed docstring only like the ones coming from ivar fields
            msg = 'Existing docstring'
            if self.docstring_lineno:
                msg += f' at line {self.docstring_lineno}'
            msg += ' is overriden'
            self.report(msg, 'docstring', lineno_offset=lineno-self.docstring_lineno)
        self.docstring = doc
        self.docstring_lineno = lineno
        # Due to the current process for parsing doc strings, some objects might already have a parsed_docstring populated at this moment. 
        # This is an unfortunate behaviour but it’s too big of a refactor for now (see https://github.com/twisted/pydoctor/issues/798).
        if self.parsed_docstring:
            self.parsed_docstring = None

    def setLineNumber(self, lineno: LineFromDocstringField | LineFromAst | int) -> None:
        """
        Save the linenumber of this object.

        If the linenumber is already set from a ast analysis, this is an no-op.
        If the linenumber is already set from docstring fields and the new linenumber
        if not from docstring fields as well, the old docstring based linumber will be replaced
        with the one from ast analysis since this takes precedence.

        @param lineno: The linenumber. 
            If the given linenumber is simply an L{int} we'll assume it's coming from the ast builder 
            and it will be converted to an L{LineFromAst} instance.
        """
        if not self.linenumber or (
            isinstance(self.linenumber, LineFromDocstringField) 
                and not isinstance(lineno, LineFromDocstringField)):
            if not isinstance(lineno, (LineFromAst, LineFromDocstringField)):
                lineno = LineFromAst(lineno)
            self.linenumber = lineno
            parent = self.parentMod
            if parent is not None:
                module_source_href = parent.source_href
                if module_source_href:
                    self.source_href = self.system.options.htmlsourcetemplate.format(
                        mod_source_href=module_source_href,
                        lineno=str(lineno)
                    )

    @property
    def description(self) -> str:
        """A string describing our source location to the user.

        If this module's code was read from a file, this returns
        its file path. In other cases, such as during unit testing,
        the full module name is returned.
        """
        source_path = self.source_path
        return self.module.fullName() if source_path is None else str(source_path)

    @property
    def page_object(self) -> 'Documentable':
        """The documentable to which the page we're documented on belongs.
        For example methods are documented on the page of their class,
        functions are documented in their module's page etc.
        """
        location = self.documentation_location
        if location is DocLocation.OWN_PAGE:
            return self
        elif location is DocLocation.PARENT_PAGE:
            parent = self.parent
            assert parent is not None
            return parent
        else:
            assert False, location

    @property
    def url(self) -> str:
        """Relative URL at which the documentation for this Documentable
        can be found.

        For page objects this method MUST return an C{.html} filename without a
        URI fragment (because L{pydoctor.templatewriter.writer.TemplateWriter}
        uses it directly to determine the output filename).
        """
        page_obj = self.page_object
        if list(self.system.root_names) == [page_obj.fullName()]:
            page_url = 'index.html'
        else:
            page_url = f'{quote(page_obj.fullName())}.html'
        if page_obj is self:
            return page_url
        else:
            return f'{page_url}#{quote(self.name)}'

    def fullName(self) -> str:
        parent = self.parent
        if parent is None:
            return self.name
        else:
            return f'{parent.fullName()}.{self.name}'

    def __repr__(self) -> str:
        return f"{self.__class__.__name__} {self.fullName()!r}"

    def docsources(self) -> Iterator['Documentable']:
        """Objects that can be considered as a source of documentation.

        The motivating example for having multiple sources is looking at a
        superclass' implementation of a method for documentation for a
        subclass'.
        """
        yield self


    def reparent(self, new_parent: 'Module', new_name: str) -> None:
        # this code attempts to preserve "rather a lot" of
        # invariants assumed by various bits of pydoctor
        # and that are of course not written down anywhere
        # :/
        self._handle_reparenting_pre()
        old_parent = self.parent
        assert isinstance(old_parent, CanContainImportsDocumentable)
        old_name = self.name
        self.parent = self.parentMod = new_parent
        self.name = new_name
        self._handle_reparenting_post()
        del old_parent.contents[old_name]
        old_parent._localNameToFullName_map[old_name] = self.fullName()
        new_parent.contents[new_name] = self
        self._handle_reparenting_post()

    def _handle_reparenting_pre(self) -> None:
        del self.system.allobjects[self.fullName()]
        for o in self.contents.values():
            o._handle_reparenting_pre()

    def _handle_reparenting_post(self) -> None:
        self.system.allobjects[self.fullName()] = self
        for o in self.contents.values():
            o._handle_reparenting_post()
    
    def _localNameToFullName(self, name: str) -> str:
        raise NotImplementedError(self._localNameToFullName)
    
    def isNameDefined(self, name:str) -> bool:
        """
        Is the given name defined in the globals/locals of self-context?
        Only the first name of a dotted name is checked.

        Returns True iff the given name can be loaded without raising `NameError`.
        """
        raise NotImplementedError(self.isNameDefined)

    def expandName(self, name: str) -> str:
        """Return a fully qualified name for the possibly-dotted `name`.

        To explain what this means, consider the following modules:

        mod1.py::

            from external_location import External
            class Local:
                pass

        mod2.py::

            from mod1 import External as RenamedExternal
            import mod1 as renamed_mod
            class E:
                pass

        In the context of mod2.E, expandName("RenamedExternal") should be
        "external_location.External" and expandName("renamed_mod.Local")
        should be "mod1.Local". """
        parts = name.split('.')
        obj: Documentable = self
        for i, p in enumerate(parts):
            full_name = obj._localNameToFullName(p)
            if full_name == p and i != 0:
                # The local name was not found.
                # If we're looking at a class, we try our luck with the inherited members
                if isinstance(obj, Class):
                    inherited = obj.find(p)
                    if inherited: 
                        full_name = inherited.fullName()
                if full_name == p:
                    # We don't have a full name
                    # TODO: Instead of returning the input, _localNameToFullName()
                    #       should probably either return None or raise LookupError.
                    full_name = f'{obj.fullName()}.{p}'
                    break
            nxt = self.system.objForFullName(full_name)
            if nxt is None:
                break
            obj = nxt
        return '.'.join([full_name] + parts[i + 1:])

    def expandAnnotationName(self, name: str) -> str:
        """
        Like L{expandName} but gives precedence to the module scope when a 
        name is defined both in the current scope and the module scope.
        """
        if self.module.isNameDefined(name):
            return self.module.expandName(name)
        elif self.isNameDefined(name):
            return self.expandName(name)
        return self.module.expandName(name)

    def resolveName(self, name: str) -> Optional['Documentable']:
        """Return the object named by "name" (using Python's lookup rules) in
        this context, if any is known to pydoctor."""
        return self.system.objForFullName(self.expandName(name))

    @property
    def privacyClass(self) -> PrivacyClass:
        """How visible this object should be."""
        return self.system.privacyClass(self)

    @property
    def isVisible(self) -> bool:
        """Is this object so private as to be not shown at all?

        This is just a simple helper which defers to self.privacyClass.
        """
        isVisible = self.privacyClass is not PrivacyClass.HIDDEN
        # If a module/package/class is hidden, all it's members are hidden as well.
        if isVisible and self.parent:
            isVisible = self.parent.isVisible
        return isVisible

    @property
    def isPrivate(self) -> bool:
        """Is this object considered private API?

        This is just a simple helper which defers to self.privacyClass.
        """
        return self.privacyClass is not PrivacyClass.PUBLIC

    @property
    def module(self) -> 'Module':
        """This object's L{Module}.

        For modules, this returns the object itself, otherwise
        the module containing the object is returned.
        """
        parentMod = self.parentMod
        assert parentMod is not None
        return parentMod

    def report(self, descr: str, section: str = 'parsing', lineno_offset: int = 0, thresh:int=-1) -> None:
        """
        Log an error or warning about this documentable object.

        @param descr: The error/warning string
        @param section: What the warning is about.
        @param lineno_offset: Offset
        @param thresh: Thresh to pass to L{System.msg}, it will use C{-1} by default, 
          meaning it will count as a violation and will fail the build if option C{-W} is passed.
          But this behaviour is not applicable if C{thresh} is greater or equal to zero.
        """

        linenumber: object
        if section in ('docstring', 'resolve_identifier_xref'):
            linenumber = self.docstring_lineno or self.linenumber
        else:
            linenumber = self.linenumber
        if linenumber:
            linenumber += lineno_offset
        elif lineno_offset and self.module is self:
            linenumber = lineno_offset
        else:
            linenumber = '???'

        self.system.msg(
            section,
            f'{self.description}:{linenumber}: {descr}',
            thresh=thresh)

    @property
    def docstring_linker(self) -> 'linker.DocstringLinker':
        """
        Returns an instance of L{DocstringLinker} suitable for resolving names
        in the context of the object. 
        """
        if self._linker is not None:
            return self._linker
        self._linker = linker._EpydocLinker(self)
        return self._linker


class CanContainImportsDocumentable(Documentable):
    def setup(self) -> None:
        super().setup()
        self._localNameToFullName_map: Dict[str, str] = {}
    
    def isNameDefined(self, name: str) -> bool:
        name = name.split('.')[0]
        if name in self.contents:
            return True
        if name in self._localNameToFullName_map:
            return True
        if not isinstance(self, Module):
            return self.module.isNameDefined(name)
        else:
            return False
    
    def localNames(self) -> Iterator[str]:
        return chain(self.contents.keys(),
                     self._localNameToFullName_map.keys())

class Module(CanContainImportsDocumentable):
    kind = DocumentableKind.MODULE
    state = ProcessingState.UNPROCESSED

    @property
    def privacyClass(self) -> PrivacyClass:
        if self.name == '__main__':
            return PrivacyClass.PRIVATE
        else:
            return super().privacyClass

    def setup(self) -> None:
        super().setup()

        self._is_c_module = False
        """Whether this module is a C-extension."""
        self._py_mod: Optional[types.ModuleType] = None
        """The live module if the module was built from introspection."""
        self._py_string: Optional[str] = None
        """The module string if the module was built from text."""

        self.all: Optional[Collection[str]] = None
        """Names listed in the C{__all__} variable of this module.

        These names are considered to be exported by the module,
        both for the purpose of C{from <module> import *} and
        for the purpose of publishing names from private modules.

        If no C{__all__} variable was found in the module, or its
        contents could not be parsed, this is L{None}.
        """

        self._docformat: Optional[str] = None

    def _localNameToFullName(self, name: str) -> str:
        if name in self.contents:
            o: Documentable = self.contents[name]
            return o.fullName()
        elif name in self._localNameToFullName_map:
            return self._localNameToFullName_map[name]
        else:
            return name

    @property
    def module(self) -> 'Module':
        return self

    @property
    def docformat(self) -> Optional[str]:
        """The name of the format to be used for parsing docstrings in this module.
        
        The docformat value are inherited from packages if a C{__docformat__} variable 
        is defined in the C{__init__.py} file.

        If no C{__docformat__} variable was found or its
        contents could not be parsed, this is L{None}.
        """
        if self._docformat:
            return self._docformat
        elif isinstance(self.parent, Package):
            return self.parent.docformat
        return None
    
    @docformat.setter
    def docformat(self, value: str) -> None:
        self._docformat = value

    def submodules(self) -> Iterator['Module']:
        """Returns an iterator over the visible submodules."""
        return (m for m in self.contents.values()
                if isinstance(m, Module) and m.isVisible)

class Package(Module):
    kind = DocumentableKind.PACKAGE

    def __repr__(self) -> str:
        return f"{self.kind.name.replace('_', ' ').title()} {self.fullName()!r}"

    # Support for namespace packages: 
    def setup(self) -> None:
        super().setup()
        self.source_paths = [self.source_path] if self.source_path else []
        self.source_hrefs: list[str] = []


# List of exceptions class names in the standard library, Python 3.8.10
_STD_LIB_EXCEPTIONS = ('ArithmeticError', 'AssertionError', 'AttributeError', 
    'BaseException', 'BlockingIOError', 'BrokenPipeError', 
    'BufferError', 'BytesWarning', 'ChildProcessError', 
    'ConnectionAbortedError', 'ConnectionError', 
    'ConnectionRefusedError', 'ConnectionResetError', 
    'DeprecationWarning', 'EOFError', 
    'EnvironmentError', 'Exception', 'FileExistsError', 
    'FileNotFoundError', 'FloatingPointError', 'FutureWarning', 
    'GeneratorExit', 'IOError', 'ImportError', 'ImportWarning', 
    'IndentationError', 'IndexError', 'InterruptedError', 
    'IsADirectoryError', 'KeyError', 'KeyboardInterrupt', 'LookupError', 
    'MemoryError', 'ModuleNotFoundError', 'NameError', 
    'NotADirectoryError', 'NotImplementedError', 
    'OSError', 'OverflowError', 'PendingDeprecationWarning', 'PermissionError', 
    'ProcessLookupError', 'RecursionError', 'ReferenceError', 
    'ResourceWarning', 'RuntimeError', 'RuntimeWarning', 'StopAsyncIteration', 
    'StopIteration', 'SyntaxError', 'SyntaxWarning', 'SystemError', 
    'SystemExit', 'TabError', 'TimeoutError', 'TypeError', 
    'UnboundLocalError', 'UnicodeDecodeError', 'UnicodeEncodeError', 
    'UnicodeError', 'UnicodeTranslateError', 'UnicodeWarning', 'UserWarning', 
    'ValueError', 'Warning', 'ZeroDivisionError')
def is_exception(cls: 'Class') -> bool:
    """
    Whether is class should be considered as 
    an exception and be marked with the special 
    kind L{DocumentableKind.EXCEPTION}.
    """
    for base in cls.mro(True, False):
        if base in _STD_LIB_EXCEPTIONS:
            return True
    return False

def topsort(graph: Mapping[Any, Sequence[T]]) -> Iterable[T]:
    """
    Given a mapping where each key-value pair correspond to a node and it's
    predecessors, return the topological order of the nodes.

    This is a simple wrapper for L{graphlib.TopologicalSorter.static_order}.
    """
    return TopologicalSorter(graph).static_order()

ClassOrStr: TypeAlias = 'Class | str'

class ClassHierarchyFinalizer:
    """
    Encapsulate code related to class hierarchies post-processing.
    """

    @classmethod
    def _init_finalbaseobjects(cls, o: Class, path:list[Class] | None = None) -> None:
        """
        The base objects are computed in two passes, first the ast visitor sets C{_initialbaseobjects}, 
        then we set C{_finalbaseobjects} from this function which should be called during post-processing.
        """
        if not path:
            path = []
        if o in path:
            cycle_str = " -> ".join([c.fullName() for c in path[path.index(o):] + [o]])
            raise mro.LinearizationError(f"Cycle found while computing inheritance hierarchy: {cycle_str}")
        path.append(o)
        if o._finalbaseobjects is not None:
            # we already computed these, so skip.
            return
        if o.rawbases:
            finalbaseobjects: List[Optional[Class]] = []
            finalbases: List[str] = []
            for i,((str_base, _), base) in enumerate(zip(o.rawbases, o._initialbaseobjects)):
                if base:
                    finalbaseobjects.append(base)
                    finalbases.append(base.fullName())
                else:
                    # Only re-resolve the base object if the base was None.
                    resolved_base = o.parent.resolveName(str_base)
                    if isinstance(resolved_base, Class):
                        base = resolved_base
                        finalbaseobjects.append(base)
                        finalbases.append(base.fullName())
                    else:
                        # the base object could not be resolved
                        finalbaseobjects.append(None)
                        finalbases.append(o._initialbases[i])
                if base:
                    # Recurse on super classes
                    cls._init_finalbaseobjects(base, path.copy())
            o._finalbaseobjects = finalbaseobjects
            o._finalbases = finalbases

    @staticmethod
    def _getbases(o: Class) -> Iterator[ClassOrStr]:
        """
        Like L{Class.baseobjects} but fallback to the expanded 
        name if the base is not resolved to a L{Class} object.
        """
        for s,b in zip(o.bases, o.baseobjects):
            if isinstance(b, Class):
                yield b
            else:
                yield s
    
    def __init__(self, classes: Iterable[Class]) -> None:
        """
        @param classes: All classes of the system.
        """
        # this calls _init_finalbaseobjects for every class and 
        # create the graph object for the ones that did not raised
        # a cycle-error.
        self.graph: dict[Class, list[ClassOrStr]]  = {}
        self.computed_mros: dict[ClassOrStr, list[ClassOrStr]] = {}
        
        for cls in classes:
            try:
                self._init_finalbaseobjects(cls)
            except mro.LinearizationError as e:
                # Set the MRO right away in case of cycles. 
                # They should not be in the graph though!
                cls.report(str(e), 'mro')
                self.computed_mros[cls] = cls._mro = list(cls.allbases(True))
            else:
                self.graph[cls] = bases = []
                for b in self._getbases(cls):
                    bases.append(b)
                    # strings are not part of the graph
                    # because they always have the same MRO: empty list.
                    
    def compute_mros(self) -> None:
        
        # If this raises a CycleError, our code is boggus since we already
        # checked for cycles ourself.
        static_order = topsort(self.graph)
        
        for cls in static_order:
            if cls in self.computed_mros or isinstance(cls, str):
                # If it's already computed, it means it's bogus
                continue
            self.computed_mros[cls] = cls._mro = self._compute_mro(cls)

    def _compute_mro(self, cls: Class) -> list[ClassOrStr]:
        """
        Compute the method resolution order for this class.
        This assumes that the MRO of the bases of the class 
        have already been computed and stored in C{self.computed_mros}.
        """
        assert cls in self.graph, f"{cls} is not known"
        
        result: list[ClassOrStr] = [cls]

        if not (bases:=self.graph[cls]):
            return result
        
        # since we compute all MRO in topological order, we can safely assume
        # that self.computed_mros contains all the MROs of the bases of this class.
        bases_mros = [self.computed_mros.get(kls, []) for kls in bases]

        # handle multiple typing.Generic case, 
        # see https://github.com/twisted/pydoctor/issues/846.
        # support documenting typing.py module by using allobject.get.
        generic = cls.system.allobjects.get(_d:='typing.Generic', _d)
        if generic in bases and any(generic in _mro for _mro in bases_mros):
            # this is safe since we checked 'generic in bases'.
            bases.remove(generic) # type: ignore[arg-type]
        
        try:
            return result + mro.c3_merge(*bases_mros, bases)
        
        except mro.LinearizationError as e:
            cls.report(f'{e} of {cls.fullName()!r}', 'mro')
            return list(cls.allbases(True))
    

def _find_dunder_constructor(cls:'Class') -> Optional['Function']:
    """
    Find the a non-default python-powered dunder constructor.
    Returns C{None} if neither C{__new__} or C{__init__} are defined.

    @note: C{__new__} takes precedence orver C{__init__}. 
        More infos: U{https://docs.python.org/3/reference/datamodel.html#object.__new__}
    """
    _new = cls.find('__new__')
    if isinstance(_new, Function):
        return _new
    elif _new is None:
        _init = cls.find('__init__')
        if isinstance(_init, Function):
            return _init
    return None

def get_constructors(cls:Class) -> Iterator[Function]:
    """
    Look for python language powered constructors or classmethod constructors.
    A constructor MUST be a method accessible in the locals of the class.
    """
    # Look for python language powered constructors.
    # If __new__ is defined, then it takes precedence over __init__
    # Blind spot: we don't understand when a Class is using a metaclass that overrides __call__.
    dunder_constructor = _find_dunder_constructor(cls)
    if dunder_constructor:
        yield dunder_constructor

    # Then look for staticmethod/classmethod constructors,
    # This only happens at the local scope level (i.e not looking in super-classes).
    for fun in cls.contents.values():
        if not isinstance(fun, Function):
            continue
        # Only static methods and class methods can be recognized as constructors
        if not fun.kind in (DocumentableKind.STATIC_METHOD, DocumentableKind.CLASS_METHOD):
            continue
        # get return annotation, if it returns the same type as self, it's a constructor method.
        if not 'return' in fun.annotations:
            # we currently only support constructor detection trought explicit annotations.
            continue 

        # annotation should be resolved at the module scope
        return_ann = astutils.node2fullname(fun.annotations['return'], cls.module)

        # pydoctor understand explicit annotation as well as the Self-Type.
        if return_ann == cls.fullName() or \
            return_ann in ('typing.Self', 'typing_extensions.Self'):
            yield fun

class Class(CanContainImportsDocumentable):
    kind = DocumentableKind.CLASS
    parent: CanContainImportsDocumentable
    decorators: Sequence[Tuple[str, Optional[Sequence[ast.expr]]]]

    # set in post-processing:
    _finalbaseobjects: Optional[List[Optional['Class']]] = None 
    _finalbases: Optional[List[str]] = None
    _mro: Optional[Sequence[Class | str]] = None

    def setup(self) -> None:
        super().setup()
        self.rawbases: Sequence[Tuple[str, ast.expr]] = []
        self.raw_decorators: Sequence[ast.expr] = []
        self.subclasses: List[Class] = []
        self._initialbases: List[str] = []
        self._initialbaseobjects: List[Optional['Class']] = []
    
    @overload
    def mro(self, include_external:'Literal[True]', include_self:bool=True) -> Sequence[Class | str]:...
    @overload
    def mro(self, include_external:'Literal[False]'=False, include_self:bool=True) -> Sequence['Class']:...
    def mro(self, include_external:bool=False, include_self:bool=True) -> Sequence[Class | str]:
        """
        Get the method resution order of this class. 

        @note: The actual correct value is only set in post-processing, if L{mro()} is called
            in the AST visitors, it will return the same as C{list(self.allbases(include_self))}.
        """
        if self._mro is None:
            return list(self.allbases(include_self))
        _mro: Sequence[Union[str, Class]]
        if include_external is False:
            _mro = [o for o in self._mro if not isinstance(o, str)]
        else:
            _mro = self._mro
        if include_self is False:
            _mro = _mro[1:]
        return _mro

    @property
    def bases(self) -> List[str]:
        """
        Fully qualified names of the bases of this class.
        """
        return self._finalbases if \
            self._finalbases is not None else self._initialbases

    
    @property
    def baseobjects(self) -> List[Optional['Class']]:
        """
        Base objects, L{None} value is inserted when the base class could not be found in the system.
        
        @note: This property is currently computed two times, a first time when we're visiting the ClassDef and initially creating the object. 
            It's computed another time in post-processing to try to resolve the names that could not be resolved the first time. This is needed when there are import cycles. 
            
            Meaning depending on the state of the system, this property can return either the initial objects or the final objects
        """
        return self._finalbaseobjects if \
            self._finalbaseobjects is not None else self._initialbaseobjects
    
    @property
    def public_constructors(self) -> Sequence['Function']:
        """
        The public constructors of this class.
        A public constructor must not be hidden and have
        arguments or have a docstring.
        """
        r = []
        for c in get_constructors(self):
            if not c.isVisible:
                continue
            args = list(c.annotations)
            try: args.remove('return')
            except ValueError: pass
            if c.kind in (DocumentableKind.CLASS_METHOD, 
                          DocumentableKind.METHOD):
                try:
                    args.pop(0)
                except IndexError:
                    pass
            if (len(args)==0 and get_docstring(c)[0] is None and 
                c.name in ('__init__', '__new__')):
                continue
            r.append(c)
        return r

    def allbases(self, include_self: bool = False) -> Iterator['Class']:
        """
        Iterate on all base objects of this class and it's super classes. Doesn't comply with MRO.
        """
        if include_self:
            yield self
        for b in self.baseobjects:
            if b is not None:
                yield from b.allbases(True)

    def find(self, name: str) -> Optional[Documentable]:
        """Look up a name in this class and its base classes.

        @return: the object with the given name, or L{None} if there isn't one
        """
        for base in self.mro():
            obj: Optional[Documentable] = base.contents.get(name)
            if obj is not None:
                return obj
        return None

    def _localNameToFullName(self, name: str) -> str:
        if name in self.contents:
            o: Documentable = self.contents[name]
            return o.fullName()
        elif name in self._localNameToFullName_map:
            return self._localNameToFullName_map[name]
        else:
            return self.parent._localNameToFullName(name)

    @property
    def constructor_params(self) -> Mapping[str, Optional[ast.expr]]:
        """A mapping of constructor parameter names to their type annotation.
        If a parameter is not annotated, its value is L{None}.
        """

        # We assume that the constructor parameters are the same as the
        # __new__()/__init__() parameters. This is incorrect if the metaclass
        # __call__() have different parameters or __init__/__new__ is using
        # signature changing decorators.
        constructor = _find_dunder_constructor(self)
        if constructor is not None:
            return constructor.annotations
        else:
            return {}


class Inheritable(Documentable):
    documentation_location = DocLocation.PARENT_PAGE

    parent: CanContainImportsDocumentable

    def docsources(self) -> Iterator[Documentable]:
        yield self
        if not isinstance(self.parent, Class):
            return
        for b in self.parent.mro(include_self=False):
            if self.name in b.contents:
                yield b.contents[self.name]

    def _localNameToFullName(self, name: str) -> str:
        return self.parent._localNameToFullName(name)
    
    def isNameDefined(self, name: str) -> bool:
        return self.parent.isNameDefined(name)

class Function(Inheritable):
    kind = DocumentableKind.FUNCTION
    is_async: bool
    annotations: Mapping[str, ast.expr | None]
    decorators: Sequence[ast.expr] | None
    signature: Signature | None
    overloads: List[FunctionOverload]

    parsed_signature: ParsedDocstring | None = None # set in get_parsed_signature()

    def setup(self) -> None:
        super().setup()
        if isinstance(self.parent, Class):
            self.kind = DocumentableKind.METHOD
        self.signature = None
        self.overloads = []

@attr.s(auto_attribs=True)
class FunctionOverload:
    """
    @note: This is not an actual documentable type. 
    """
    primary: Function
    signature: Signature | None 
    decorators: Sequence[ast.expr]
    parsed_signature: ParsedDocstring | None = None # set in get_parsed_signature()

class Attribute(Inheritable):
    kind: Optional[DocumentableKind] = DocumentableKind.ATTRIBUTE
    annotation: Optional[ast.expr] = None
    decorators: Optional[Sequence[ast.expr]] = None
    value: Optional[ast.expr] = None
    """
    The value of the assignment expression. 

    None value means the value is not initialized at the current point of the the process. 
    """

# Work around the attributes of the same name within the System class.
_ModuleT = Module
_PackageT = Package

def import_mod_from_file_location(module_full_name:str, path: Path) -> types.ModuleType:
    spec = importlib.util.spec_from_file_location(module_full_name, path)
    if spec is None: 
        raise RuntimeError(f"Cannot find spec for module {module_full_name} at {path}")
    py_mod = importlib.util.module_from_spec(spec)
    loader = spec.loader
    assert isinstance(loader, importlib.abc.Loader), loader
    loader.exec_module(py_mod)
    return py_mod


# Declare the types that we consider as functions (also when they are coming
# from a C extension)
func_types: Tuple[Type[Any], ...] = (types.BuiltinFunctionType, types.FunctionType)
if hasattr(types, "MethodDescriptorType"):
    # This is Python >= 3.7 only
    func_types += (types.MethodDescriptorType, )
else:
    func_types += (type(str.join), )
if hasattr(types, "ClassMethodDescriptorType"):
    # This is Python >= 3.7 only
    func_types += (types.ClassMethodDescriptorType, )
else:
    func_types += (type(dict.__dict__["fromkeys"]), )

class ModuleNotAdded(Exception):
    ...

_default_extensions = object()
class System:
    """A collection of related documentable objects.

    PyDoctor documents collections of objects, often the contents of a
    package.
    """

    systemBuilder: Type[ISystemBuilder]

    # Not assigned here for circularity reasons:
    defaultBuilder: Type[ASTBuilder]
    syntaxTreeParser: Type[SyntaxTreeParser]

    extensions: List[str] = cast('List[str]', _default_extensions)
    """
    List of extensions.

    By default, all built-in pydoctor extensions will be loaded.
    Override this value to cherry-pick extensions. 
    """

    custom_extensions: List[str] = []
    """
    Additional list of extensions to load alongside default extensions.
    """

    show_attr_value = (DocumentableKind.CONSTANT, 
                       DocumentableKind.TYPE_VARIABLE, 
                       DocumentableKind.TYPE_ALIAS)
    """
    What kind of attributes we should display the value for?
    """

    def __init__(self, options: Optional['Options'] = None):
        self.allobjects: Dict[str, Documentable] = {}
        self.rootobjects: List[_ModuleT] = []

        self.violations = 0
        """The number of docstring problems found.
        This is used to determine whether to fail the build when using
        the --warnings-as-errors option, so it should only be increased
        for problems that the user can fix.
        """

        if options:
            self.options: Options = options
        else:
            self.options = Options.defaults()
            self.options.verbosity = 3

        self.projectname = 'my project'

        self.parse_errors: Dict[str, Set[str]] = defaultdict(set)
        """
        Dict from the name of the thing we're rendering (C{section}) to the FullNames of objects for which the rendereable elements failed to parse.
        Typically the renderable element is the C{docstring}, but it can be the decorators, parameter default values or any other colorized AST.
        """

        self.verboselevel = 0
        self.needsnl = False
        self.once_msgs: Set[Tuple[str, str]] = set()

        # We're using the id() of the modules as key, and not the fullName becaue modules can
        # be reparented, generating KeyError.
        self.unprocessed_modules: List[_ModuleT] = []

        self.module_count = 0
        self.processing_modules: List[str] = []
        self.buildtime = datetime.datetime.now()
        self.intersphinx = SphinxInventory(logger=self.msg)
        self._ast_parser = self.syntaxTreeParser()

        # Since privacy handling now uses fnmatch, we cache results so we don't re-run matches all the time.
        # We use the fullName of the objets as the dict key in order to bind a full name to a privacy, not an object to a privacy.
        # this way, we are sure the objects' privacy stay true even if we reparent them manually.
        self._privacyClassCache: Dict[str, PrivacyClass] = {}
        
        # workaround cyclic import issue
        from pydoctor import extensions

        # Initialize the extension system
        self._factory = factory.Factory()
        self._astbuilder_visitors: List[Type['astutils.NodeVisitorExt']] = []
        self._post_processor = extensions.PriorityProcessor(self)
        
        if self.extensions == _default_extensions:
            self.extensions = list(extensions.get_extensions())
        assert isinstance(self.extensions, list)
        assert isinstance(self.custom_extensions, list)
        # pydoctor.astbuilder includes some required extensions, so always add it.
        self.extensions = ['pydoctor.astbuilder'] + self.extensions
        for ext in self.extensions + self.custom_extensions:
            # Load extensions
            extensions.load_extension_module(self, ext)

    @property
    def Class(self) -> Type['Class']:
        return self._factory.Class
    @property
    def Function(self) -> Type['Function']:
        return self._factory.Function
    @property
    def Module(self) -> Type['Module']:
        return self._factory.Module
    @property
    def Package(self) -> Type['Package']:
        return self._factory.Package
    @property
    def Attribute(self) -> Type['Attribute']:
        return self._factory.Attribute

    @property
    def sourcebase(self) -> Optional[str]:
        return self.options.htmlsourcebase

    @property
    def root_names(self) -> Collection[str]:
        """The top-level package/module names in this system."""
        return {obj.name for obj in self.rootobjects}

    def progress(self, section: str, i: int, n: Optional[int], msg: str) -> None:
        if n is None:
            d = str(i)
        else:
            d = f'{i}/{n}'
        if self.options.verbosity == 0 and sys.stdout.isatty():
            print('\r'+d, msg, end='')
            sys.stdout.flush()
            if d == n:
                self.needsnl = False
                print()
            else:
                self.needsnl = True

    def msg(self,
            section: str,
            msg: str,
            thresh: int = 0,
            topthresh: int = 100,
            nonl: bool = False,
            wantsnl: bool = True,
            once: bool = False
            ) -> None:
        """
        Log a message. pydoctor's logging system is bit messy.
        
        @param section: API doc generation step this message belongs to.
        @param msg: The message.
        @param thresh: The minimum verbosity level of the system for this message to actually be printed.
            Meaning passing thresh=-1 will make message still display if C{-q} is passed but not if C{-qq}. 
            Similarly, passing thresh=1 will make the message only apprear if the verbosity level is at least increased once with C{-v}.
            Using negative thresh will count this message as a violation and will fail the build if option C{-W} is passed.
        @param topthresh: The maximum verbosity level of the system for this message to actually be printed.
        """
        if once:
            if (section, msg) in self.once_msgs:
                return
            else:
                self.once_msgs.add((section, msg))

        if thresh < 0:
            # Apidoc build messages are generated using negative threshold
            # and we have separate reporting for them,
            # on top of the logging system.
            self.violations += 1

        if thresh <= self.options.verbosity <= topthresh:
            if self.needsnl and wantsnl:
                print()
            print(msg, end='')
            if nonl:
                self.needsnl = True
                sys.stdout.flush()
            else:
                self.needsnl = False
                print('')

    def objForFullName(self, fullName: str) -> Optional[Documentable]:
        return self.allobjects.get(fullName)

    def find_object(self, full_name: str) -> Optional[Documentable]:
        """Look up an object using a potentially outdated full name.

        A name can become outdated if the object is reparented:
        L{objForFullName()} will only be able to find it under its new name,
        but we might still have references to the old name.

        @param full_name: The fully qualified name of the object.
        @return: The object, or L{None} if the name is external (it does not
            match any of the roots of this system).
        @raise LookupError: If the object is not found, while its name does
            match one of the roots of this system.
        """
        obj = self.objForFullName(full_name)
        if obj is not None:
            return obj

        # The object might have been reparented, in which case there will
        # be an alias at the original location; look for it using expandName().
        name_parts = full_name.split('.', 1)
        for root_obj in self.rootobjects:
            if root_obj.name == name_parts[0]:
                obj = self.objForFullName(root_obj.expandName(name_parts[1]))
                if obj is not None:
                    return obj
                raise LookupError(full_name)

        return None

    def objectsOfType(self, cls: Union[Type['DocumentableT'], str]) -> Iterator['DocumentableT']:
        """Iterate over all instances of C{cls} present in the system. """
        if isinstance(cls, str):
            cls = utils.findClassFromDottedName(cls, 'objectsOfType', 
                base_class=cast(Type['DocumentableT'], Documentable))
        assert isinstance(cls, type)
        for o in self.allobjects.values():
            if isinstance(o, cls):
                yield o

    def privacyClass(self, ob: Documentable) -> PrivacyClass:
        ob_fullName = ob.fullName()
        cached_privacy = self._privacyClassCache.get(ob_fullName)
        if cached_privacy is not None:
            return cached_privacy
        
        # kind should not be None, this is probably a relica of a past age of pydoctor.
        # but keep it just in case.
        if ob.kind is None:
            return PrivacyClass.HIDDEN
        
        privacy = PrivacyClass.PUBLIC
        if ob.name.startswith('_') and \
               not (ob.name.startswith('__') and ob.name.endswith('__')):
            privacy = PrivacyClass.PRIVATE
        
        # Precedence order: CLI arguments order
        # Check exact matches first, then qnmatch
        _found_exact_match = False
        for priv, match in reversed(self.options.privacy):
            if ob_fullName == match:
                privacy = priv
                _found_exact_match = True
                break
        if not _found_exact_match:
            for priv, match in reversed(self.options.privacy):
                if qnmatch.qnmatch(ob_fullName, match):
                    privacy = priv
                    break

        # Store in cache
        self._privacyClassCache[ob_fullName] = privacy
        return privacy

    def membersOrder(self, ob: Documentable) -> Callable[[Documentable], Tuple[Any, ...]]:
        """
        Returns a callable suitable to be used with L{sorted} function. 
        Used to sort the given object's members for presentation.

        Users can customize class and module members order independently, or can override this method
        with a custom system class for further tweaks.
        """
        from pydoctor.templatewriter.util import objects_order
        if isinstance(ob, Class):
            return objects_order(self.options.cls_member_order)
        else:
            return objects_order(self.options.mod_member_order)
           
    def addObject(self, obj: Documentable) -> None:
        """Add C{object} to the system."""

        if obj.parent:
            obj.parent.contents[obj.name] = obj
        elif isinstance(obj, _ModuleT):
            self.rootobjects.append(obj)
        else:
            raise ValueError(f'Top-level object is not a module: {obj!r}')

        first = self.allobjects.setdefault(obj.fullName(), obj)
        if obj is not first:
            self.handleDuplicate(obj)

    # if we assume:
    #
    # - that svn://divmod.org/trunk is checked out into ~/src/Divmod
    #
    # - that http://divmod.org/trac/browser/trunk is the trac URL to the
    #   above directory
    #
    # - that ~/src/Divmod/Nevow/nevow is passed to pydoctor as an argument
    #
    # we want to work out the sourceHref for nevow.flat.ten.  the answer
    # is http://divmod.org/trac/browser/trunk/Nevow/nevow/flat/ten.py.
    #
    # we can work this out by finding that Divmod is the top of the svn
    # checkout, and posixpath.join-ing the parts of the filePath that
    # follows that.
    #
    #  http://divmod.org/trac/browser/trunk
    #                          ~/src/Divmod/Nevow/nevow/flat/ten.py

    def _formatSourceHref(self, source_path: Path) -> str | None:
        if self.sourcebase is None:
            return None
        projBaseDir = self.options.projectbasedirectory
        assert projBaseDir is not None
        try:
            relative = source_path.relative_to(projBaseDir).as_posix()
        except ValueError:
            # The links cannot be computed because the source path lies outside base directory.
            return None
        
        return f'{self.sourcebase}/{relative}'

    def setSourceHref(self, mod: _ModuleT, source_path: Path) -> None:
        # Note: this is used only for module, for other objects
        # it's done in setLineNumber().

        # pydoctor supports generating documentation covering more than one package, 
        # in which case it is not certain that all of the source is even viewable below a single URL.
        # We ignore this limitation by not assigning sourceHref for now, but it would be good to add support for it.
        # a way to do it would be to match package names to different source bases and project base directories. 
        # --html-viewsource-base=zope.interface:https://github.com/zopefoundation/zope.interface/tree/master
        # --html-viewsource-base=zope.component:https://github.com/zopefoundation/zope.component/tree/master
        # --project-base-dir=zope.interface:./zope.interface/
        # --project-base-dir=zope.component:./zope.component/
        # ./zope.interface/src/zope ./zope.component/src/zope
        
        if href:=self._formatSourceHref(source_path):
            if mod.source_href is None:
                mod.source_href = href
            # Support for namespace packages: their location can be split off
            # several distributions, needing several hrefs.
            if mod.kind is DocumentableKind.NAMESPACE_PACKAGE:
                assert isinstance(mod, Package)
                mod.source_hrefs.append(href)

    @overload
    def _addPackageOrModule(self,
            modpath: Path,
            modname: str,
            parent: Optional[_PackageT],
            is_package: Literal[False] = False,
            is_namespace_package: Literal[False] = False,
            ) -> _ModuleT: ...

    @overload
    def _addPackageOrModule(self,
            modpath: Path,
            modname: str,
            parent: Optional[_PackageT],
            is_package: Literal[True], 
            is_namespace_package: bool = False, 
            ) -> _PackageT: ...

    def _addPackageOrModule(self,
            modpath: Path,
            modname: str,
            parent: Optional[_PackageT] = None,
            is_package: bool = False, 
            is_namespace_package: bool = False,
            ) -> _ModuleT:
        
        """
        Add a single package or module from path.

        @returns: The added module or package. In the case of namespace packages, 
            the existing package will be returned if it's found.
        @raise ModuleNotAdded: If the module has been discarded because a module under the same 
            name already Exist.
        """
        mod: Module | Package
        if is_namespace_package and isinstance(maybe_mod := self.allobjects.get(
            f'{parent.fullName()}.{modname}' if parent else modname), Package) and \
            maybe_mod.kind is DocumentableKind.NAMESPACE_PACKAGE:
            # A namespace package already exist for this package name, then
            # simply add a new source path to it, calling setSourceHref() will
            # append a new ref
            mod = maybe_mod
            mod.source_paths.append(modpath)

        else:
            cls = self.Package if is_package else self.Module
            # Create the new module
            mod = cls(self, modname, parent, modpath)
            if is_namespace_package:
                mod.kind = DocumentableKind.NAMESPACE_PACKAGE
            if not self._addUnprocessedModule(mod):
                raise ModuleNotAdded
        
        self.setSourceHref(mod, modpath)
        return mod

    def _addUnprocessedModule(self, mod: _ModuleT) -> bool:
        """
        First add the new module into the unprocessed_modules list. 
        Handle eventual duplication of module names, and finally add the 
        module to the system.

        @returns: Whether the module has been sucessfully added.
        """
        assert mod.state is ProcessingState.UNPROCESSED
        first = self.allobjects.get(mod.fullName())
        if first is not None:
            # At this step of processing only modules exists
            assert isinstance(first, Module)
            return self._handleDuplicateModule(first, mod)

        self.unprocessed_modules.append(mod)
        self.addObject(mod)
        self.progress(
            "analyzeModule", len(self.allobjects),
            None, "modules and packages discovered")        
        self.module_count += 1
        return True

    def _handleDuplicateModule(self, first: _ModuleT, dup: _ModuleT) -> bool:
        """
        This is called when two modules have the same name. 

        Current rules are the following: 
            - C-modules wins over regular python modules
            - Packages wins over modules
            - Namespace Packages wins over regular packages
            - Else, the last added module wins
        """
        if first._is_c_module and not isinstance(dup, Package):
            # C-modules wins
            dup.report(f"discarding duplicate {str(dup)} because existing C extension has the same name", thresh=0)
            return False
        elif isinstance(first, Package) and not isinstance(dup, Package):
            # Packages wins over module
            dup.report(f"discarding duplicate {str(dup)} because existing package has the same name", thresh=0)
            return False
        elif first.kind is DocumentableKind.NAMESPACE_PACKAGE and dup.kind is not DocumentableKind.NAMESPACE_PACKAGE:
            # Namespace packages wins over regular package
            dup.report(f"discarding duplicate {str(dup)} because existing namespace package has the same name", thresh=0)
            return False

        # Else, the last added module wins
        dup.report(f"discarding existing {str(first)} because {str(dup)} overrides it", thresh=0)
        self._remove(first)
        self.unprocessed_modules.remove(first)
        return self._addUnprocessedModule(dup)

    def _introspectThing(self, thing: object, parent: CanContainImportsDocumentable, parentMod: _ModuleT) -> None:
        for k, v in thing.__dict__.items():
            if (isinstance(v, func_types)
                    # In PyPy 7.3.1, functions from extensions are not
                    # instances of the abstract types in func_types, it will have the type 'builtin_function_or_method'.
                    # Additionnaly cython3 produces function of type 'cython_function_or_method', 
                    # so se use a heuristic on the class name as a fall back detection.
                    or (hasattr(v, "__class__") and 
                        v.__class__.__name__.endswith('function_or_method'))):
                f = self.Function(self, k, parent)
                f.parentMod = parentMod
                f.docstring = v.__doc__
                f.decorators = None
                try:
                    f.signature = signature(v)
                except ValueError:
                    # function has an invalid signature.
                    parent.report(f"Cannot parse signature of {parent.fullName()}.{k}")
                    f.signature = None
                except TypeError:
                    # in pypy we get a TypeError calling signature() on classmethods, 
                    # because apparently, they are not callable :/
                    f.signature = None
                        
                f.is_async = False
                f.annotations = {name: None for name in f.signature.parameters} if f.signature else {}
                self.addObject(f)
            elif isinstance(v, type):
                c = self.Class(self, k, parent)
                c.rawbases = []
                c.parentMod = parentMod
                c.docstring = v.__doc__
                self.addObject(c)
                self._introspectThing(v, c, parentMod)

    def introspectModule(self,
            path: Path,
            module_name: str,
            package: Optional[_PackageT]
            ) -> _ModuleT:

        if package is None:
            module_full_name = module_name
        else:
            module_full_name = f'{package.fullName()}.{module_name}'

        py_mod = import_mod_from_file_location(module_full_name, path)
        is_package = py_mod.__package__ == py_mod.__name__

        factory = self.Package if is_package else self.Module
        module = factory(self, module_name, package, path)
        
        module.docstring = py_mod.__doc__
        module._is_c_module = True
        module._py_mod = py_mod
        
        self._addUnprocessedModule(module)
        return module
    
    def addPackage(self, package_path: Path, parent: Optional[_PackageT] = None, 
                   is_namespace_package: bool = False) -> None:

        if not is_namespace_package:
            pkg_source_path = package_path / '__init__.py'
        else:
            pkg_source_path = package_path
        
        try:
            package = self._addPackageOrModule(pkg_source_path, package_path.name, 
                                        parent, is_package=True, 
                                        is_namespace_package=is_namespace_package)
        except ModuleNotAdded:
            # a message is logged in case of clashing modules.
            return 
        
        for path in sorted(package_path.iterdir()):
            if path.is_dir():
                if is_namespace_package:
                    # A namespace package should only be nested under other nspackages
                    # if it's nested under a regular package, then we ignore it since 
                    # the chances are this is not part of the API. 
                    # What if we have an empty namespace package? Then nohting special happens
                    self.addPackage(path, package, (self._is_pep420_namespace_package(path) or 
                                                    self._is_oldschool_namespace_package(path)))
                elif not self._is_pep420_namespace_package(path):
                    self.addPackage(path, package)

            elif path.name != '__init__.py' and not path.name.startswith('.'):
                self.addModule(path, package)

    def addModule(self, path: Path, parent: Optional[_PackageT]) -> None:
        name = path.name
        for suffix in importlib.machinery.all_suffixes():
            if not name.endswith(suffix):
                continue
            module_name = name[:-len(suffix)]
            if suffix in importlib.machinery.EXTENSION_SUFFIXES:
                if self.options.introspect_c_modules:
                    self.introspectModule(path, module_name, parent)
            elif suffix in importlib.machinery.SOURCE_SUFFIXES:
                try:
                    self._addPackageOrModule(path, module_name, parent)
                except ModuleNotAdded: 
                    pass
            break

    def _is_pep420_namespace_package(self, path: Path) -> bool:
        return not path.joinpath('__init__.py').is_file()
    
    def _is_oldschool_namespace_package(self, path: Path) -> bool:
        try:
            tree = self._ast_parser.parseFile(
                path.joinpath('__init__.py'))
        except Exception:
            return False
        return astutils.is_old_school_namespace_package(tree)

    
    def _remove(self, o: Documentable) -> None:
        del self.allobjects[o.fullName()]
        oc = list(o.contents.values())
        for c in oc:
            self._remove(c)

    def handleDuplicate(self, obj: Documentable) -> None:
        """
        This is called when we see two objects with the same
        .fullName(), for example::

            class C:
                if something:
                    def meth(self):
                        implementation 1
                else:
                    def meth(self):
                        implementation 2

        The default is that the second definition "wins".
        """
        i = 0
        fullName = obj.fullName()
        while (fullName + ' ' + str(i)) in self.allobjects:
            i += 1
        prev = self.allobjects[fullName]
        obj.report(f"duplicate {str(prev)}", thresh=1)
        self._remove(prev)
        prev.name = obj.name + ' ' + str(i)
        def readd(o: Documentable) -> None:
            self.allobjects[o.fullName()] = o
            for c in o.contents.values():
                readd(c)
        readd(prev)
        self.allobjects[fullName] = obj


    def getProcessedModule(self, modname: str) -> Optional[_ModuleT]:
        mod = self.allobjects.get(modname)
        if mod is None:
            return None
        if not isinstance(mod, Module):
            return None

        if mod.state is ProcessingState.UNPROCESSED:
            self.processModule(mod)

        assert mod.state in (ProcessingState.PROCESSING, ProcessingState.PROCESSED), mod.state
        return mod

    def processModule(self, mod: _ModuleT) -> None:
        assert mod.state is ProcessingState.UNPROCESSED
        assert mod in self.unprocessed_modules
        mod.state = ProcessingState.PROCESSING
        self.unprocessed_modules.remove(mod)
        if mod.source_path is None:
            assert mod._py_string is not None
        if mod._is_c_module:
            self.processing_modules.append(mod.fullName())
            self.msg("processModule", "processing %s"%(self.processing_modules), 1)
            self._introspectThing(mod._py_mod, mod, mod)
            mod.state = ProcessingState.PROCESSED
            head = self.processing_modules.pop()
            assert head == mod.fullName()
        else:
            builder = self.defaultBuilder(self)
            ast = None
            if mod._py_string is not None:
                ast = builder.parseString(mod._py_string, mod)
            elif mod.kind is not DocumentableKind.NAMESPACE_PACKAGE:
                # There is no AST for namespace packages.
                assert mod.source_path is not None
                ast = builder.parseFile(mod.source_path, mod)
            if ast:
                self.processing_modules.append(mod.fullName())
                if mod._py_string is None:
                    self.msg("processModule", "processing %s"%(self.processing_modules), 1)
                builder.processModuleAST(ast, mod)
                mod.state = ProcessingState.PROCESSED
                head = self.processing_modules.pop()
                assert head == mod.fullName()
        self.progress(
            'process',
            self.module_count - len(self.unprocessed_modules),
            self.module_count,
            f"modules processed, {self.violations} warnings")


    def process(self) -> None:
        while self.unprocessed_modules:
            mod = next(iter(self.unprocessed_modules))
            self.processModule(mod)
        self.postProcess()


    def postProcess(self) -> None:
        """Called when there are no more unprocessed modules.

        Analysis of relations between documentables can be done here,
        without the risk of drawing incorrect conclusions because modules
        were not fully processed yet.

        @See: L{extensions.PriorityProcessor}.
        """
        self._post_processor.apply_processors()

    def fetchIntersphinxInventories(self, cache: CacheT) -> None:
        """
        Download and parse intersphinx inventories based on configuration.
        """
        for url in self.options.intersphinx:
            self.intersphinx.update(cache, url)

def defaultPostProcess(system:'System') -> None:

    # Init class graph and final bases.
    class_finalizer = ClassHierarchyFinalizer(
        system.objectsOfType(Class))
    
    # Compute MROs
    class_finalizer.compute_mros()

    for cls in system.objectsOfType(Class):
        # Compute subclasses
        for b in cls.baseobjects:
            if b is not None:
                b.subclasses.append(cls)

        # Checking whether the class is an exception
        if is_exception(cls):
            cls.kind = DocumentableKind.EXCEPTION
            
    for attrib in system.objectsOfType(Attribute):
       _inherits_instance_variable_kind(attrib)

    from pydoctor import epydoc2stan
    for ns in system.objectsOfType(Package):
        if ns.kind is not DocumentableKind.NAMESPACE_PACKAGE:
            continue
        # This overrides the namespace package docstring, 
        # but since they are not supposed to include
        # any documentation this will not be an issue.
        ns.docstring = epydoc2stan.get_namespace_docstring(ns)

def _inherits_instance_variable_kind(attr: Attribute) -> None:
    """
    If any of the inherited members of a class variable is an instance variable,
    then the subclass' class variable become an instance variable as well.
    """
    if attr.kind is not DocumentableKind.CLASS_VARIABLE:
        return
    docsources = attr.docsources()
    next(docsources)
    for inherited in docsources:
        if inherited.kind is DocumentableKind.INSTANCE_VARIABLE:
            attr.kind = DocumentableKind.INSTANCE_VARIABLE
            break

def get_docstring(
        obj: Documentable
        ) -> Tuple[Optional[str], Optional[Documentable]]:
    """
    Fetch the docstring for a documentable.
    Treat empty docstring as undocumented.

    :returns:
        - C{(docstring, source)} if the object is documented.
        - C{(None, None)} if the object has no docstring (even inherited).
        - C{(None, source)} if the object has an empty docstring.
    """
    for source in obj.docsources():
        doc = source.docstring
        if doc:
            return doc, source
        if doc is not None:
            # Treat empty docstring as undocumented.
            return None, source
    return None, None


class SystemBuildingError(Exception):
    """
    Raised when there is a (handled) fatal error while adding modules to the builder.
    """

class ISystemBuilder(abc.ABC):
    """
    Interface class for building a system.
    """
    @abc.abstractmethod
    def __init__(self, system: 'System') -> None:
        """
        Create the builder.
        """
    @abc.abstractmethod
    def addModule(self, path: Path, parent_name: Optional[str] = None, ) -> None:
        """
        Add a module or package from file system path to the pydoctor system. 
        If the path points to a directory, adds all submodules recursively.

        @raises SystemBuildingError: If there is an error while adding the module/package.
        """
    @abc.abstractmethod
    def addModuleString(self, text: str, modname: str,
                        parent_name: Optional[str] = None,
                        is_package: bool = False, ) -> None:
        """
        Add a module from text to the system.
        """
    @abc.abstractmethod
    def buildModules(self) -> None:
        """
        Build the modules.
        """

class SystemBuilder(ISystemBuilder):
    """
    This class is only an adapter for some System methods related to module building. 
    """
    def __init__(self, system: 'System') -> None:
        self.system = system
        self._added: Set[Path] = set()

    def addModule(self, path: Path, parent_name: Optional[str] = None, ) -> None:
        if path in self._added:
            return
        # Path validity check
        projBaseDir = self.system.options.projectbasedirectory
        if projBaseDir is not None:
            # Note: Path.is_relative_to() was only added in Python 3.9,
            #       so we have to use this workaround for now.
            try:
                path.relative_to(projBaseDir)
            except ValueError:
                if self.system.options.htmlsourcebase:  
                    # We now support building documentation when the source path is outside of the build directory.
                    # We simply leave a warning and skip the sourceHref attribute.
                    # https://github.com/twisted/pydoctor/issues/658
                    # https://github.com/twisted/pydoctor/issues/870
                    _warn_msg = f"No source links can be generated for module {path}: source path lies outside base directory {projBaseDir}"
                    self.system.msg('addPackage', _warn_msg, once=True)
        parent: Optional[Package] = None
        if parent_name:
            _p = self.system.allobjects[parent_name]
            assert isinstance(_p, Package)
            parent = _p
        if path.is_dir():
            self.system.msg('addPackage', f"adding directory {path}")
            self.system.addPackage(path, parent, 
                                   self.system._is_pep420_namespace_package(path) or 
                                   self.system._is_oldschool_namespace_package(path))
        elif path.is_file():
            self.system.msg('addModule', f"adding module {path}")
            self.system.addModule(path, parent)
        elif path.exists():
            raise SystemBuildingError(f"Source path is neither file nor directory: {path}")
        else:
            raise SystemBuildingError(f"Source path does not exist: {path}")
        self._added.add(path)

    def addModuleString(self, text: str, modname: str,
                        parent_name: Optional[str] = None,
                        is_package: bool = False, ) -> None:
        if parent_name is None:
            parent = None
        else:
            # Set containing package as parent.
            parent = self.system.allobjects[parent_name]
            assert isinstance(parent, Package), f"{parent.fullName()} is not a Package, it's a {parent.kind}"
        
        factory = self.system.Package if is_package else self.system.Module
        mod = factory(self.system, name=modname, parent=parent, source_path=None)
        mod._py_string = textwrap.dedent(text)
        self.system._addUnprocessedModule(mod)

    def buildModules(self) -> None:
        self.system.process()

System.systemBuilder = SystemBuilder

def prepend_package(builderT:Type[ISystemBuilder], package:str) -> Type[ISystemBuilder]:
    """
    Get a new system builder class, that extends the original C{builder} such that it will always use a "fake" 
    C{package} to be the only root object of the system and add new modules under it.
    """
    
    class PrependPackageBuidler(builderT): # type:ignore
        """
        Support for option C{--prepend-package}.
        """

        def __init__(self, system: 'System', *, package:str) -> None:
            super().__init__(system)
            
            self.package = package
            
            prependedpackage = None
            for m in package.split('.'):
                prependedpackage = system.Package(
                    system, m, prependedpackage)
                system.addObject(prependedpackage)
        
        def addModule(self, path: Path, parent_name: Optional[str] = None, ) -> None:
            if parent_name is None:
                parent_name = self.package
            super().addModule(path, parent_name)
        
        def addModuleString(self, text: str, modname: str,
                            parent_name: Optional[str] = None,
                            is_package: bool = False, ) -> None:
            if parent_name is None:
                parent_name = self.package
            super().addModuleString(text, modname, parent_name, is_package=is_package)
    
    return utils.partialclass(PrependPackageBuidler, package=package)



from contextlib import redirect_stdout
from io import StringIO
from pathlib import Path
import re
import sys

import pytest

from pydoctor.options import Options
from pydoctor import driver

from . import CapSys


def geterrtext(*options: str) -> str:
    """
    Run CLI with options and return the output triggered by system exit.
    """
    se = sys.stderr
    f = StringIO()
    print(options)
    sys.stderr = f
    try:
        try:
            driver.main(list(options))
        except SystemExit:
            pass
        else:
            assert False, "did not fail"
    finally:
        sys.stderr = se
    return f.getvalue()

def test_invalid_option() -> None:
    err = geterrtext('--no-such-option')
    assert 'unrecognized arguments: --no-such-option' in err

def test_cannot_advance_blank_system() -> None:
    err = geterrtext('--make-html')
    assert 'No source paths given' in err

def test_no_systemclasses_py3() -> None:
    err = geterrtext('--system-class')
    assert 'expected one argument' in err

def test_invalid_systemclasses() -> None:
    err = geterrtext('--system-class=notdotted')
    assert 'dotted name' in err
    err = geterrtext('--system-class=no-such-module.System')
    assert 'could not import module' in err
    err = geterrtext('--system-class=pydoctor.model.Class')
    assert 'is not a subclass' in err


def test_projectbasedir_absolute(tmp_path: Path) -> None:
    """
    The --project-base-dir option, when given an absolute path, should set that
    path as the projectbasedirectory attribute on the options object.

    Previous versions of this test tried using non-existing paths and compared
    the string representations, but that was unreliable, since the input path
    might contain a symlink that will be resolved, such as "/home" on macOS.
    Using L{Path.samefile()} is reliable, but requires an existing path.
    """
    assert tmp_path.is_absolute()
    options = Options.from_args(["--project-base-dir", str(tmp_path)])
    assert options.projectbasedirectory is not None
    assert options.projectbasedirectory.samefile(tmp_path)
    assert options.projectbasedirectory.is_absolute()


@pytest.mark.skipif("platform.python_implementation() == 'PyPy' and platform.system() == 'Windows'")
def test_projectbasedir_symlink(tmp_path: Path) -> None:
    """
    The --project-base-dir option, when given a path containing a symbolic link,
    should resolve the path to the target directory.
    """
    target = tmp_path / 'target'
    target.mkdir()
    link = tmp_path / 'link'
    link.symlink_to('target', target_is_directory=True)
    assert link.samefile(target)

    options = Options.from_args(["--project-base-dir", str(link)])
    assert options.projectbasedirectory is not None
    assert options.projectbasedirectory.samefile(target)
    assert options.projectbasedirectory.is_absolute()


def test_projectbasedir_relative() -> None:
    """
    The --project-base-dir option, when given a relative path, should convert
    that path to absolute and set it as the projectbasedirectory attribute on
    the options object.
    """
    relative = "projbasedirvalue"
    options = Options.from_args(["--project-base-dir", relative])
    assert options.projectbasedirectory is not None
    assert options.projectbasedirectory.is_absolute()
    assert options.projectbasedirectory.name == relative
    assert options.projectbasedirectory.parent == Path.cwd()


def test_help_option(capsys: CapSys) -> None:
    """
    pydoctor --help 
    """
    try:
        driver.main(args=['--help'])
    except SystemExit:
        assert '--project-name PROJECTNAME' in capsys.readouterr().out
    else:
        assert False

def test_cache_enabled_by_default() -> None:
    """
    Intersphinx object caching is enabled by default.
    """
    options = Options.defaults()
    assert options.enable_intersphinx_cache


def test_cli_warnings_on_error() -> None:
    """
    The --warnings-as-errors option is disabled by default.
    This is the test for the long form of the CLI option.
    """
    options = Options.defaults()
    assert options.warnings_as_errors == False

    options = Options.from_args(['--warnings-as-errors'])
    assert options.warnings_as_errors == True


def test_project_version_default() -> None:
    """
    When no --project-version is provided, it will default empty string.
    """
    options = Options.defaults()
    assert options.projectversion == ''


def test_project_version_string() -> None:
    """
    --project-version can be passed as a simple string.
    """
    options = Options.from_args(['--project-version', '1.2.3.rc1'])
    assert options.projectversion == '1.2.3.rc1'


def test_main_project_name_guess(capsys: CapSys) -> None:
    """
    When no project name is provided in the CLI arguments, a default name
    is used and logged.
    """
    exit_code = driver.main(args=[
        '-v', '--testing',
        'pydoctor/test/testpackages/basic/'
        ])

    assert exit_code == 0
    assert "Guessing 'basic' for project name." in capsys.readouterr().out


def test_main_project_name_option(capsys: CapSys) -> None:
    """
    When a project name is provided in the CLI arguments nothing is logged.
    """
    exit_code = driver.main(args=[
        '-v', '--testing',
        '--project-name=some-name',
        'pydoctor/test/testpackages/basic/'
        ])

    assert exit_code == 0
    assert 'Guessing ' not in capsys.readouterr().out


def test_main_return_zero_on_warnings() -> None:
    """
    By default it will return 0 as exit code even when there are warnings.
    """
    stream = StringIO()
    with redirect_stdout(stream):
        exit_code = driver.main(args=[
            '--html-writer=pydoctor.test.InMemoryWriter',
            'pydoctor/test/testpackages/report_trigger/'
            ])

    assert exit_code == 0
    assert "__init__.py:8: Unknown field 'bad_field'" in stream.getvalue()
    assert 'report_module.py:9: Cannot find link target for "BadLink"' in stream.getvalue()


def test_main_return_non_zero_on_warnings() -> None:
    """
    When `-W` is used it returns 3 as exit code when there are warnings.
    """
    stream = StringIO()
    with redirect_stdout(stream):
        exit_code = driver.main(args=[
            '-W',
            '--html-writer=pydoctor.test.InMemoryWriter',
            'pydoctor/test/testpackages/report_trigger/'
            ])

    assert exit_code == 3
    assert "__init__.py:8: Unknown field 'bad_field'" in stream.getvalue()
    assert 'report_module.py:9: Cannot find link target for "BadLink"' in stream.getvalue()


@pytest.mark.skipif("platform.python_implementation() == 'PyPy' and platform.system() == 'Windows'")
def test_main_symlinked_paths(tmp_path: Path) -> None:
    """
    The project base directory and package/module directories are normalized
    in the same way, such that System.setSourceHref() can call Path.relative_to()
    on them.
    """
    link = tmp_path / 'src'
    link.symlink_to(Path.cwd(), target_is_directory=True)

    exit_code = driver.main(args=[
        '--project-base-dir=.',
        '--html-viewsource-base=http://example.com',
        f'{link}/pydoctor/test/testpackages/basic/'
        ])
    assert exit_code == 0


def test_main_source_outside_basedir(capsys: CapSys) -> None:
    """
    If a --project-base-dir is given, all package and module paths should
    be located inside that base directory if source links wants to be generated.
    Otherwise it's OK, but no source links will be genrated
    """
    assert driver.main(args=[
        '--html-viewsource-base=notnone',
        '--project-base-dir=docs',
        'pydoctor/test/testpackages/basic/'
        ]) == 0
    re.match("No source links can be generated for module .+/pydoctor/test/testpackages/basic/: source path lies outside base directory .+/docs\n", 
        capsys.readouterr().out)

    assert driver.main(args=[
        '--project-base-dir=docs',
        'pydoctor/test/testpackages/basic/'
        ]) == 0
    assert "No source links can be generated" not in capsys.readouterr().out

    assert driver.main(args=[
        '--html-viewsource-base=notnone',
        '--project-base-dir=pydoctor/test/testpackages/',
        'pydoctor/test/testpackages/basic/'
        ]) == 0
    assert "No source links can be generated" not in capsys.readouterr().out


def test_make_intersphix(tmp_path: Path) -> None:
    """
    --make-intersphinx without --make-html will only produce the Sphinx inventory object.

    This is also an integration test for the Sphinx inventory writer.
    """
    inventory = tmp_path / 'objects.inv'
    exit_code = driver.main(args=[
        '--project-base-dir=.',
        '--make-intersphinx',
        '--project-name=acme-lib',
        '--project-version=20.12.0-dev123',
        '--html-output', str(tmp_path),
        'pydoctor/test/testpackages/basic/'
        ])

    assert exit_code == 0
    # No other files are created, other than the inventory.
    assert [p.name for p in tmp_path.iterdir()] == ['objects.inv']
    assert inventory.is_file()
    assert b'Project: acme-lib\n# Version: 20.12.0-dev123\n' in inventory.read_bytes()

def test_index_symlink(tmp_path: Path) -> None:
    """
    Test that the default behaviour is to create symlinks, at least on unix.

    For windows users, this has not been a success, so we automatically fallback to copying the file now.
    See https://github.com/twisted/pydoctor/issues/808, https://github.com/twisted/pydoctor/issues/720.
    """
    import platform
    exit_code = driver.main(args=['--html-output', str(tmp_path), 'pydoctor/test/testpackages/basic/'])
    assert exit_code == 0
    link = (tmp_path / 'basic.html')
    assert link.exists()
    if platform.system() == 'Windows':
        assert link.is_symlink() or link.is_file()
    else:
        assert link.is_symlink()

def test_index_hardlink(tmp_path: Path) -> None:
    """
    Test for option --use-hardlink wich enforce the usage of harlinks.
    """
    exit_code = driver.main(args=['--use-hardlink', '--html-output', str(tmp_path), 'pydoctor/test/testpackages/basic/'])
    assert exit_code == 0
    assert (tmp_path / 'basic.html').exists()
    assert not (tmp_path / 'basic.html').is_symlink()
    assert (tmp_path / 'basic.html').is_file()


def test_apidocs_help(tmp_path: Path) -> None:
    """
    Checks that the help page is well generated.
    """
    exit_code = driver.main(args=['--html-output', str(tmp_path), 'pydoctor/test/testpackages/basic/'])
    assert exit_code == 0
    help_page = (tmp_path / 'apidocs-help.html').read_text()
    assert '>Search</h2>' in help_page

def test_htmlbaseurl_option_all_pages(tmp_path: Path) -> None:
    """
    Check that the canonical link is included in all html pages, including summary pages.
    """
    exit_code = driver.main(args=[
        '--html-base-url=https://example.com.abcde',
        '--html-output', str(tmp_path), 'pydoctor/test/testpackages/basic/'])
    assert exit_code == 0
    for t in tmp_path.iterdir():
        if not t.name.endswith('.html'):
            continue
        filename = t.name
        if t.stem == 'basic':
            filename = 'index.html' # since we have only one module it's linked as index.html
        assert f'<link rel="canonical" href="https://example.com.abcde/{filename}"' in t.read_text(encoding='utf-8')
    




"""
Parser for google-style docstrings. 

@See: L{pydoctor.epydoc.markup.numpy}
@See: L{pydoctor.epydoc.markup._napoleon}
"""
from __future__ import annotations

from pydoctor.epydoc.markup import ObjClass, ParserFunction
from pydoctor.epydoc.markup._napoleon import NapoelonDocstringParser


def get_parser(objclass: ObjClass | None) -> ParserFunction:
    """
    Returns the parser function. Behaviour will depend on the documentable type and system options.
    """
    return NapoelonDocstringParser(objclass).parse_google_docstring




from typing import Iterable
from pydoctor.test import CapSys
from pydoctor.test.epydoc.test_restructuredtext import parse_rst
from pydoctor import visitor
from docutils import nodes

def dump(node: nodes.Node, text:str='') -> None:
    print('{}{:<15} line: {}, rawsource: {}'.format(
        text,
        type(node).__name__, 
        node.line,
        getattr(node, 'rawsource', node.astext()).replace('\n', '\\n')))  

class DocutilsNodeVisitor(visitor.Visitor[nodes.Node]):
    def unknown_visit(self, ob: nodes.Node) -> None:
        pass
    def unknown_departure(self, ob: nodes.Node) -> None:
        pass
    
    @classmethod
    def get_children(cls, ob:nodes.Node) -> Iterable[nodes.Node]:
        if isinstance(ob, nodes.Element):
            return ob.children
        return []

class MainVisitor(DocutilsNodeVisitor):
    def visit_title_reference(self, node: nodes.Node) -> None:
        raise self.SkipNode()

class ParagraphDump(visitor.VisitorExt[nodes.Node]):
    when = visitor.When.AFTER
    def visit_paragraph(self, node: nodes.Node) -> None:
        dump(node)

class TitleReferenceDumpAfter(visitor.VisitorExt[nodes.Node]):
    when = visitor.When.AFTER
    def visit_title_reference(self, node: nodes.Node) -> None:
        dump(node)

class GenericDump(DocutilsNodeVisitor):
    def unknown_visit(self, node: nodes.Node) -> None:
        dump(node, '[visit-main] ')
    def unknown_departure(self, node: nodes.Node) -> None:
        dump(node, '[depart-main] ')

class GenericDumpAfter(visitor.VisitorExt[nodes.Node]):
    when = visitor.When.INNER
    def unknown_visit(self, node: nodes.Node) -> None:
        dump(node, '[visit-inner] ')
    def unknown_departure(self, node: nodes.Node) -> None:
        dump(node, '[depart-inner] ')

class GenericDumpBefore(visitor.VisitorExt[nodes.Node]):
    when = visitor.When.OUTTER
    def unknown_visit(self, node: nodes.Node) -> None:
        dump(node, '[visit-outter] ')
    def unknown_departure(self, node: nodes.Node) -> None:
        dump(node, '[depart-outter] ')


def test_visitor_ext(capsys:CapSys) -> None:

    parsed_doc = parse_rst('''
Hello
=====

Dolor sit amet
''')
    doc = parsed_doc.to_node()

    vis = GenericDump()
    vis.extensions.add(GenericDumpAfter, GenericDumpBefore)
    vis.walkabout(doc)
    assert capsys.readouterr().out == r'''[visit-outter] document        line: None, rawsource: 
[visit-main] document        line: None, rawsource: 
[visit-inner] document        line: None, rawsource: 
[visit-outter] title           line: 3, rawsource: Hello
[visit-main] title           line: 3, rawsource: Hello
[visit-inner] title           line: 3, rawsource: Hello
[visit-outter] Text            line: None, rawsource: Hello
[visit-main] Text            line: None, rawsource: Hello
[visit-inner] Text            line: None, rawsource: Hello
[depart-inner] Text            line: None, rawsource: Hello
[depart-main] Text            line: None, rawsource: Hello
[depart-outter] Text            line: None, rawsource: Hello
[depart-inner] title           line: 3, rawsource: Hello
[depart-main] title           line: 3, rawsource: Hello
[depart-outter] title           line: 3, rawsource: Hello
[visit-outter] paragraph       line: 5, rawsource: Dolor sit amet
[visit-main] paragraph       line: 5, rawsource: Dolor sit amet
[visit-inner] paragraph       line: 5, rawsource: Dolor sit amet
[visit-outter] Text            line: None, rawsource: Dolor sit amet
[visit-main] Text            line: None, rawsource: Dolor sit amet
[visit-inner] Text            line: None, rawsource: Dolor sit amet
[depart-inner] Text            line: None, rawsource: Dolor sit amet
[depart-main] Text            line: None, rawsource: Dolor sit amet
[depart-outter] Text            line: None, rawsource: Dolor sit amet
[depart-inner] paragraph       line: 5, rawsource: Dolor sit amet
[depart-main] paragraph       line: 5, rawsource: Dolor sit amet
[depart-outter] paragraph       line: 5, rawsource: Dolor sit amet
[depart-inner] document        line: None, rawsource: 
[depart-main] document        line: None, rawsource: 
[depart-outter] document        line: None, rawsource: 
'''


def test_visitor(capsys:CapSys) -> None:

    parsed_doc = parse_rst('''
Fizz
====
Lorem ipsum `notfound`.

Buzz
****

Lorem ``ipsum``

.. code-block:: python

   x = 0

.. note::

   Dolor sit amet
   `notfound`.

   .. code-block:: python

      y = 1

Dolor sit amet `another link <notfound>`.
Dolor sit amet `link <notfound>`.
bla blab balba.
''')
    doc = parsed_doc.to_node()

    MainVisitor(visitor.ExtList(TitleReferenceDumpAfter)).walkabout(doc)
    assert capsys.readouterr().out == r'''title_reference line: None, rawsource: `notfound`
title_reference line: None, rawsource: `notfound`
title_reference line: None, rawsource: `another link <notfound>`
title_reference line: None, rawsource: `link <notfound>`
'''
    
    vis = MainVisitor()
    vis.extensions.add(ParagraphDump, TitleReferenceDumpAfter)
    vis.walkabout(doc)
    assert capsys.readouterr().out == r'''paragraph       line: 4, rawsource: Lorem ipsum `notfound`.
title_reference line: None, rawsource: `notfound`
paragraph       line: 9, rawsource: Lorem ``ipsum``
paragraph       line: 17, rawsource: Dolor sit amet\n`notfound`.
title_reference line: None, rawsource: `notfound`
paragraph       line: 24, rawsource: Dolor sit amet `another link <notfound>`.\nDolor sit amet `link <notfound>`.\nbla blab balba.
title_reference line: None, rawsource: `another link <notfound>`
title_reference line: None, rawsource: `link <notfound>`
'''



import ast
from typing import List, Type
from pydoctor import astutils, extensions,  model

class ModVisitor(extensions.ModuleVisitorExt):

    def depart_AnnAssign(self, node: ast.AnnAssign) -> None:
        """
        Called after an annotated assignment definition is visited.
        """
        ctx = self.visitor.builder.current
        if not isinstance(ctx, model.Class):
            # check if the current context object is a class
            return

        if not any(ctx.expandName(b) == 'pydantic.BaseModel' for b in ctx.bases):
            # check if the current context object if a class derived from ``pydantic.BaseModel``
            return

        dottedname = astutils.node2dottedname(node.target)
        if not dottedname or len(dottedname)!=1:
            # check if the assignment is a simple name, otherwise ignore it
            return
        
        # Get the attribute from current context
        attr = ctx.contents[dottedname[0]]

        assert isinstance(attr, model.Attribute)

        # All class variables that are not annotated with ClassVar will be transformed to instance variables.
        if astutils.is_using_typing_classvar(attr.annotation, attr):
            return

        if attr.kind == model.DocumentableKind.CLASS_VARIABLE:
            attr.kind = model.DocumentableKind.INSTANCE_VARIABLE

def setup_pydoctor_extension(r:extensions.ExtRegistrar) -> None:
    r.register_astbuilder_visitor(ModVisitor)

class PydanticSystem2(model.System):
    # Add our custom extension
    extensions: List[str] = []
    custom_extensions = ['pydoctor.test.test_pydantic_fields']

## Testing code

import pytest
from pydoctor.test.test_astbuilder import fromText, PydanticSystem

pydantic_systemcls_param = pytest.mark.parametrize('systemcls', (PydanticSystem, PydanticSystem2))

@pydantic_systemcls_param
def test_pydantic_fields(systemcls: Type[model.System]) -> None:
    src = '''
    from typing import ClassVar
    from pydantic import BaseModel, Field
    class Model(BaseModel):
        a: int
        b: int = Field(...)
        name:str = 'Jane Doe'
        kind:ClassVar = 'person'
    '''

    mod = fromText(src, modname='mod', systemcls=systemcls)

    assert mod.contents['Model'].contents['a'].kind == model.DocumentableKind.INSTANCE_VARIABLE
    assert mod.contents['Model'].contents['b'].kind == model.DocumentableKind.INSTANCE_VARIABLE
    assert mod.contents['Model'].contents['name'].kind == model.DocumentableKind.INSTANCE_VARIABLE
    assert mod.contents['Model'].contents['kind'].kind == model.DocumentableKind.CLASS_VARIABLE




"""
Various bits of reusable code related to L{ast.AST} node processing.
"""
from __future__ import annotations

import inspect
import sys
from numbers import Number
from typing import Any, Callable, Collection, Iterator, Optional, List, Iterable, Sequence, TYPE_CHECKING, Tuple, Union, cast
from inspect import BoundArguments, Signature
import ast

unparse = ast.unparse

from pydoctor import visitor

if TYPE_CHECKING:
    from pydoctor import model

# AST visitors

def iter_values(node: ast.AST) -> Iterator[ast.AST]:
    for _, value in ast.iter_fields(node):
        if isinstance(value, list):
            for item in value:
                if isinstance(item, ast.AST):
                    yield item
        elif isinstance(value, ast.AST):
            yield value

class NodeVisitor(visitor.PartialVisitor[ast.AST]):
    """
    Generic AST node visitor. This class does not work like L{ast.NodeVisitor}, 
    it only visits statements directly within a C{B{body}}. Also, visitor methods can't return anything.

    :See: L{visitor} for more informations.
    """
    def generic_visit(self, node: ast.AST) -> None:
        """
        Helper method to visit a node by calling C{visit()} on each child of the node. 
        This is useful because this vistitor only visits statements inside C{.body} attribute. 
        
        So if one wants to visit L{ast.Expr} children with their visitor, they should include::

            def visit_Expr(self, node:ast.Expr):
                self.generic_visit(node)
        """
        for v in iter_values(node):
            self.visit(v)
    
    @classmethod
    def get_children(cls, node: ast.AST) -> Iterable[ast.AST]:
        """
        Returns the nested nodes in the body of a node.
        """
        body: Optional[Sequence[ast.AST]] = getattr(node, 'body', None)
        if body is not None:
            for child in body:
                yield child

class NodeVisitorExt(visitor.VisitorExt[ast.AST]):
    ...

_AssingT = Union[ast.Assign, ast.AnnAssign]
def iterassign(node:_AssingT) -> Iterator[Optional[List[str]]]:
    """
    Utility function to iterate assignments targets. 

    Useful for all the following AST assignments::

        var:int=2
        self.var = target = node.astext()
        ol = ['extensions']

    NOT Useful for the following AST assignments::

        x, y = [1,2]

    Example:

    >>> from pydoctor.astutils import iterassign
    >>> from ast import parse
    >>> node = parse('self.var = target = thing[0] = node.astext()').body[0]
    >>> list(iterassign(node))
    [['self', 'var'], ['target'], None]
    """
    for target in node.targets if isinstance(node, ast.Assign) else [node.target]:
        dottedname = node2dottedname(target) 
        yield dottedname

def node2dottedname(node: Optional[ast.AST]) -> Optional[List[str]]:
    """
    Resove expression composed by L{ast.Attribute} and L{ast.Name} nodes to a list of names. 
    """
    parts = []
    while isinstance(node, ast.Attribute):
        parts.append(node.attr)
        node = node.value
    if isinstance(node, ast.Name):
        parts.append(node.id)
    else:
        return None
    parts.reverse()
    return parts

def node2fullname(expr: Optional[ast.AST], 
                  ctx: model.Documentable | None = None, 
                  *,
                  expandName:Callable[[str], str] | None = None) -> Optional[str]:
    if expandName is None:
        if ctx is None:
            raise TypeError('this function takes exactly two arguments')
        expandName = ctx.expandName
    elif ctx is not None:
        raise TypeError('this function takes exactly two arguments')

    dottedname = node2dottedname(expr)
    if dottedname is None:
        return None
    return expandName('.'.join(dottedname))

def bind_args(sig: Signature, call: ast.Call) -> BoundArguments:
    """
    Binds the arguments of a function call to that function's signature.
    @raise TypeError: If the arguments do not match the signature.
    """
    kwargs = {
        kw.arg: kw.value
        for kw in call.keywords
        # When keywords are passed using '**kwargs', the 'arg' field will
        # be None. We don't currently support keywords passed that way.
        if kw.arg is not None
        }
    return sig.bind(*call.args, **kwargs)


def get_str_value(expr:ast.expr) -> Optional[str]:
    if isinstance(expr, ast.Constant) and isinstance(expr.value, str):
        return expr.value
    return None
def get_num_value(expr:ast.expr) -> Optional[Number]:
    if isinstance(expr, ast.Constant) and isinstance(expr.value, Number):
        return expr.value
    return None
def _is_str_constant(expr: ast.expr, s: str) -> bool:
    return isinstance(expr, ast.Constant) and expr.value == s

def get_int_value(expr: ast.expr) -> Optional[int]:
    num = get_num_value(expr)
    if isinstance(num, int):
        return num # type:ignore[unreachable]
    return None

def is__name__equals__main__(cmp: ast.Compare) -> bool:
    """
    Returns whether or not the given L{ast.Compare} is equal to C{__name__ == '__main__'}.
    """
    return isinstance(cmp.left, ast.Name) \
    and cmp.left.id == '__name__' \
    and len(cmp.ops) == 1 \
    and isinstance(cmp.ops[0], ast.Eq) \
    and len(cmp.comparators) == 1 \
    and _is_str_constant(cmp.comparators[0], '__main__')

def is_using_typing_final(expr: Optional[ast.AST], 
                    ctx:'model.Documentable') -> bool:
    return is_using_annotations(expr, ("typing.Final", "typing_extensions.Final"), ctx)

def is_using_typing_classvar(expr: Optional[ast.AST], 
                    ctx:'model.Documentable') -> bool:
    return is_using_annotations(expr, ('typing.ClassVar', "typing_extensions.ClassVar"), ctx)

def is_using_annotations(expr: Optional[ast.AST], 
                            annotations:Sequence[str], 
                            ctx:'model.Documentable') -> bool:
    """
    Detect if this expr is firstly composed by one of the specified annotation(s)' full name.
    """
    full_name = node2fullname(expr, ctx)
    if full_name in annotations:
        return True
    if isinstance(expr, ast.Subscript):
        # Final[...] or typing.Final[...] expressions
        if isinstance(expr.value, (ast.Name, ast.Attribute)):
            value = expr.value
            full_name = node2fullname(value, ctx)
            if full_name in annotations:
                return True
    return False

def get_node_block(node: ast.AST) -> tuple[ast.AST, str]:
    """
    Tell in wich block the given node lives in. 
    
    A block is defined by a tuple: (parent node, fieldname)
    """
    try:
        parent = next(get_parents(node))
    except StopIteration:
        raise ValueError(f'node has no parents: {node}')
    for fieldname, value in ast.iter_fields(parent):
        if value is node or (isinstance(value, (list, tuple)) and node in value):
            break
    else:
        raise ValueError(f"node {node} not found in {parent}")
    return parent, fieldname

def get_assign_docstring_node(assign:ast.Assign | ast.AnnAssign) -> Str | None:
    """
    Get the docstring for a L{ast.Assign} or L{ast.AnnAssign} node.

    This helper function relies on the non-standard C{.parent} attribute on AST nodes
    to navigate upward in the tree and determine this node direct siblings.
    """
    # if this call raises an ValueError it means that we're doing something nasty with the ast...
    parent_node, fieldname = get_node_block(assign)
    statements = getattr(parent_node, fieldname, None)
    
    if isinstance(statements, Sequence):
        # it must be a sequence if it's not None since an assignment 
        # can only be a part of a compound statement.
        assign_index = statements.index(assign)
        try:
            right_sibling = statements[assign_index+1]
        except IndexError:
            return None
        if isinstance(right_sibling, ast.Expr) and \
           get_str_value(right_sibling.value) is not None:
            return cast(Str, right_sibling.value)
    return None

def is_none_literal(node: ast.expr) -> bool:
    """Does this AST node represent the literal constant None?"""
    return isinstance(node, ast.Constant) and node.value is None
    
def unstring_annotation(node: ast.expr, ctx:'model.Documentable', section:str='annotation') -> ast.expr:
    """Replace all strings in the given expression by parsed versions.
    @return: The unstringed node. If parsing fails, an error is logged
        and the original node is returned.
    """
    try:
        expr = _AnnotationStringParser().visit(node)
    except SyntaxError as ex:
        module = ctx.module
        assert module is not None
        module.report(f'syntax error in {section}: {ex}', lineno_offset=node.lineno, section=section)
        return node
    else:
        assert isinstance(expr, ast.expr), expr
        return expr

class _AnnotationStringParser(ast.NodeTransformer):
    """Implementation of L{unstring_annotation()}.

    When given an expression, the node returned by L{ast.NodeVisitor.visit()}
    will also be an expression.
    If any string literal contained in the original expression is either
    invalid Python or not a singular expression, L{SyntaxError} is raised.
    """

    def _parse_string(self, value: str) -> ast.expr:
        statements = ast.parse(value).body
        if len(statements) != 1:
            raise SyntaxError("expected expression, found multiple statements")
        stmt, = statements
        if isinstance(stmt, ast.Expr):
            # Expression wrapped in an Expr statement.
            expr = self.visit(stmt.value)
            assert isinstance(expr, ast.expr), expr
            return expr
        else:
            raise SyntaxError("expected expression, found statement")

    def visit_Subscript(self, node: ast.Subscript) -> ast.Subscript:
        value = self.visit(node.value)
        if isinstance(value, ast.Name) and value.id == 'Literal':
            # Literal[...] expression; don't unstring the arguments.
            slice = node.slice
        elif isinstance(value, ast.Attribute) and value.attr == 'Literal':
            # typing.Literal[...] expression; don't unstring the arguments.
            slice = node.slice
        else:
            # Other subscript; unstring the slice.
            slice = self.visit(node.slice)
        return ast.copy_location(ast.Subscript(value=value, slice=slice, ctx=node.ctx), node)

    def visit_fast(self, node: ast.expr) -> ast.expr:
        return node
    
    visit_Attribute = visit_Name = visit_fast

    def visit_Constant(self, node: ast.Constant) -> ast.expr:
        value = node.value
        if isinstance(value, str):
            return ast.copy_location(self._parse_string(value), node)
        else:
            const = self.generic_visit(node)
            assert isinstance(const, ast.Constant), const
            return const

def upgrade_annotation(node: ast.expr, ctx: model.Documentable, section:str='annotation') -> ast.expr:
    """
    Transform the annotation to use python 3.10+ syntax. 
    """
    return _UpgradeDeprecatedAnnotations(ctx).visit(node)

class _UpgradeDeprecatedAnnotations(ast.NodeTransformer):
    if TYPE_CHECKING:
        def visit(self, node:ast.AST) -> ast.expr:...

    def __init__(self, ctx: model.Documentable) -> None:
        def _node2fullname(node:ast.expr) -> str | None:
            return node2fullname(node, expandName=ctx.expandAnnotationName)
        self.node2fullname = _node2fullname

    def _union_args_to_bitor(self, args: list[ast.expr], ctxnode:ast.AST) -> ast.BinOp:
        assert len(args) > 1
        *others, right = args
        if len(others) == 1:
            rnode = ast.BinOp(left=others[0], right=right, op=ast.BitOr())
        else:
            rnode = ast.BinOp(left=self._union_args_to_bitor(others, ctxnode), right=right, op=ast.BitOr())
    
        return ast.fix_missing_locations(ast.copy_location(rnode, ctxnode))

    def visit_Name(self, node: ast.Name | ast.Attribute) -> Any:
        fullName = self.node2fullname(node)
        if fullName in DEPRECATED_TYPING_ALIAS_BUILTINS:
            return ast.Name(id=DEPRECATED_TYPING_ALIAS_BUILTINS[fullName], ctx=ast.Load())
        # TODO: Support all deprecated aliases including the ones in the collections.abc module.
        # In order to support that we need to generate the parsed docstring directly and include 
        # custom refmap or transform the ast such that missing imports are added.
        return node

    visit_Attribute = visit_Name

    def visit_Subscript(self, node: ast.Subscript) -> ast.expr:
        node.value = self.visit(node.value)
        node.slice = self.visit(node.slice)
        fullName = self.node2fullname(node.value)
        
        if fullName == 'typing.Union':
            # typing.Union can be used with a single type or a 
            # tuple of types, includea single element tuple, which is the same
            # as the directly using the type: Union[x] == Union[(x,)] == x
            slice_ = node.slice
            if isinstance(slice_, ast.Tuple):
                args = slice_.elts
                if len(args) > 1:
                    return self._union_args_to_bitor(args, node)
                elif len(args) == 1:
                    return args[0]
            elif isinstance(slice_, (ast.Attribute, ast.Name, ast.Subscript, ast.BinOp)):
                return slice_
        
        elif fullName == 'typing.Optional':
            # typing.Optional requires a single type, so we don't process when slice is a tuple.
            slice_ = node.slice
            if isinstance(slice_, (ast.Attribute, ast.Name, ast.Subscript, ast.BinOp)):
                return self._union_args_to_bitor([slice_, ast.Constant(value=None)], node)

        return node
    
DEPRECATED_TYPING_ALIAS_BUILTINS = {
        "typing.Text": 'str',
        "typing.Dict": 'dict',
        "typing.Tuple": 'tuple',
        "typing.Type": 'type',
        "typing.List": 'list',
        "typing.Set": 'set',
        "typing.FrozenSet": 'frozenset',
}

# These do not belong in the deprecated builtins aliases, so we make sure it doesn't happen.
assert 'typing.Union' not in DEPRECATED_TYPING_ALIAS_BUILTINS
assert 'typing.Optional' not in DEPRECATED_TYPING_ALIAS_BUILTINS

TYPING_ALIAS = (
        "typing.Hashable",
        "typing.Awaitable",
        "typing.Coroutine",
        "typing.AsyncIterable",
        "typing.AsyncIterator",
        "typing.Iterable",
        "typing.Iterator",
        "typing.Reversible",
        "typing.Sized",
        "typing.Container",
        "typing.Collection",
        "typing.Callable",
        "typing.AbstractSet",
        "typing.MutableSet",
        "typing.Mapping",
        "typing.MutableMapping",
        "typing.Sequence",
        "typing.MutableSequence",
        "typing.ByteString",
        "typing.Deque",
        "typing.MappingView",
        "typing.KeysView",
        "typing.ItemsView",
        "typing.ValuesView",
        "typing.ContextManager",
        "typing.AsyncContextManager",
        "typing.DefaultDict",
        "typing.OrderedDict",
        "typing.Counter",
        "typing.ChainMap",
        "typing.Generator",
        "typing.AsyncGenerator",
        "typing.Pattern",
        "typing.Match",
        # Special forms
        "typing.Union",
        "typing.Literal",
        "typing.Optional",
        *DEPRECATED_TYPING_ALIAS_BUILTINS, 
    )

SUBSCRIPTABLE_CLASSES_PEP585 = (
        "tuple",
        "list",
        "dict",
        "set",
        "frozenset",
        "type",
        "builtins.tuple",
        "builtins.list",
        "builtins.dict",
        "builtins.set",
        "builtins.frozenset",
        "builtins.type",
        "collections.deque",
        "collections.defaultdict",
        "collections.OrderedDict",
        "collections.Counter",
        "collections.ChainMap",
        "collections.abc.Awaitable",
        "collections.abc.Coroutine",
        "collections.abc.AsyncIterable",
        "collections.abc.AsyncIterator",
        "collections.abc.AsyncGenerator",
        "collections.abc.Iterable",
        "collections.abc.Iterator",
        "collections.abc.Generator",
        "collections.abc.Reversible",
        "collections.abc.Container",
        "collections.abc.Collection",
        "collections.abc.Callable",
        "collections.abc.Set",
        "collections.abc.MutableSet",
        "collections.abc.Mapping",
        "collections.abc.MutableMapping",
        "collections.abc.Sequence",
        "collections.abc.MutableSequence",
        "collections.abc.ByteString",
        "collections.abc.MappingView",
        "collections.abc.KeysView",
        "collections.abc.ItemsView",
        "collections.abc.ValuesView",
        "contextlib.AbstractContextManager",
        "contextlib.AbstractAsyncContextManager",
        "re.Pattern",
        "re.Match",
    )

def is_typing_annotation(node: ast.AST, ctx: 'model.Documentable') -> bool:
    """
    Whether this annotation node refers to a typing alias.
    """
    return is_using_annotations(node, TYPING_ALIAS, ctx) or \
            is_using_annotations(node, SUBSCRIPTABLE_CLASSES_PEP585, ctx)

def get_docstring_node(node: ast.AST) -> Str | None:
    """
    Return the docstring node for the given class, function or module
    or None if no docstring can be found.
    """
    if not isinstance(node, (ast.AsyncFunctionDef, ast.FunctionDef, ast.ClassDef, ast.Module)) or not node.body:
        return None
    node = node.body[0]
    if isinstance(node, ast.Expr):
        if isinstance(node.value, Str):
            return node.value
    return None

class _StrMeta(type):
    def __instancecheck__(self, instance: object) -> bool:
        if isinstance(instance, ast.expr):
            return get_str_value(instance) is not None
        return False

class Str(ast.expr, metaclass=_StrMeta):
    """
    Wraps ast.Constant/ast.Str for `isinstance` checks and annotations. 
    Ensures that the value is actually a string.
    Do not try to instanciate this class.
    """

    value: str

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        raise TypeError(f'{Str.__qualname__} cannot be instanciated')

def extract_docstring_linenum(node: Str) -> int:
    r"""
    In older CPython versions, the AST only tells us the end line
    number and we must approximate the start line number.
    This approximation is correct if the docstring does not contain
    explicit newlines ('\n') or joined lines ('\' at end of line).

    Leading blank lines are stripped by cleandoc(), so we must
    return the line number of the first non-blank line.
    """
    doc = node.value
    lineno = node.lineno

    # Leading blank lines are stripped by cleandoc(), so we must
    # return the line number of the first non-blank line.
    for ch in doc:
        if ch == '\n':
            lineno += 1
        elif not ch.isspace():
            break
    
    return lineno

def extract_docstring(node: Str) -> Tuple[int, str]:
    """
    Extract docstring information from an ast node that represents the docstring.

    @returns: 
        - The line number of the first non-blank line of the docsring. See L{extract_docstring_linenum}.
        - The docstring to be parsed, cleaned by L{inspect.cleandoc}.
    """
    value = node.value
    lineno = extract_docstring_linenum(node)
    return lineno, inspect.cleandoc(value)


def infer_type(expr: ast.expr) -> Optional[ast.expr]:
    """Infer a literal expression's type.
    @param expr: The expression's AST.
    @return: A type annotation, or None if the expression has no obvious type.
    """
    try:
        value: object = ast.literal_eval(expr)
    except (ValueError, TypeError):
        return None
    else:
        ann = _annotation_for_value(value)
        if ann is None:
            return None
        else:
            return ast.fix_missing_locations(ast.copy_location(ann, expr))

def _annotation_for_value(value: object) -> Optional[ast.expr]:
    if value is None:
        return None
    name = type(value).__name__
    if isinstance(value, (dict, list, set, tuple)):
        ann_elem = _annotation_for_elements(value)
        if isinstance(value, dict):
            ann_value = _annotation_for_elements(value.values())
            if ann_value is None:
                ann_elem = None
            elif ann_elem is not None:
                ann_elem = ast.Tuple(elts=[ann_elem, ann_value], ctx=ast.Load())
        if ann_elem is not None:
            if name == 'tuple':
                ann_elem = ast.Tuple(elts=[ann_elem, ast.Constant(value=...)], ctx=ast.Load())
            return ast.Subscript(value=ast.Name(id=name, ctx=ast.Load()),
                                 slice=ann_elem,
                                 ctx=ast.Load())
    return ast.Name(id=name, ctx=ast.Load())

def _annotation_for_elements(sequence: Iterable[object]) -> Optional[ast.expr]:
    names = set()
    for elem in sequence:
        ann = _annotation_for_value(elem)
        if isinstance(ann, ast.Name):
            names.add(ann.id)
        else:
            # Nested sequences are too complex.
            return None
    if len(names) == 1:
        name = names.pop()
        return ast.Name(id=name, ctx=ast.Load())
    else:
        # Empty sequence or no uniform type.
        return None

      
class Parentage(ast.NodeVisitor):
    """
    Add C{parent} attribute to ast nodes instances.
    """
    def __init__(self) -> None:
        self.current: ast.AST | None = None

    def generic_visit(self, node: ast.AST) -> None:
        current = self.current
        setattr(node, 'parent', current)
        self.current = node
        for child in ast.iter_child_nodes(node):
            self.generic_visit(child)
        self.current = current

def get_parents(node:ast.AST) -> Iterator[ast.AST]:
    """
    Once nodes have the C{.parent} attribute with {Parentage}, use this function
    to get a iterator on all parents of the given node up to the root module.
    """
    def _yield_parents(n:Optional[ast.AST]) -> Iterator[ast.AST]:
        if n:
            yield n
            p = cast(ast.AST, getattr(n, 'parent', None))
            yield from _yield_parents(p)
    yield from _yield_parents(getattr(node, 'parent', None))

#Part of the astor library for Python AST manipulation.
#License: 3-clause BSD
#Copyright (c) 2015 Patrick Maupin
_op_data = """
    GeneratorExp                1

          Assign                1
       AnnAssign                1
       AugAssign                0
            Expr                0
           Yield                1
       YieldFrom                0
              If                1
             For                0
        AsyncFor                0
           While                0
          Return                1

           Slice                1
       Subscript                0
           Index                1
        ExtSlice                1
    comprehension_target        1
           Tuple                0
  FormattedValue                0

           Comma                1
       NamedExpr                1
          Assert                0
           Raise                0
    call_one_arg                1

          Lambda                1
           IfExp                0

   comprehension                1
              Or   or           1
             And   and          1
             Not   not          1

              Eq   ==           1
              Gt   >            0
             GtE   >=           0
              In   in           0
              Is   is           0
           NotEq   !=           0
              Lt   <            0
             LtE   <=           0
           NotIn   not in       0
           IsNot   is not       0

           BitOr   |            1
          BitXor   ^            1
          BitAnd   &            1
          LShift   <<           1
          RShift   >>           0
             Add   +            1
             Sub   -            0
            Mult   *            1
             Div   /            0
             Mod   %            0
        FloorDiv   //           0
         MatMult   @            0
          PowRHS                1
          Invert   ~            1
            UAdd   +            0
            USub   -            0
             Pow   **           1
           Await                1
             Num                1
        Constant                1
"""

_op_data = [x.split() for x in _op_data.splitlines()] # type:ignore
_op_data = [[x[0], ' '.join(x[1:-1]), int(x[-1])] for x in _op_data if x] # type:ignore
for _index in range(1, len(_op_data)):
    _op_data[_index][2] *= 2 # type:ignore
    _op_data[_index][2] += _op_data[_index - 1][2] # type:ignore

_deprecated: Collection[str] = ()
if sys.version_info >= (3, 12):
    _deprecated = ('Num', 'Str', 'Bytes', 'Ellipsis', 'NameConstant')
_precedence_data = dict((getattr(ast, x, None), z) for x, y, z in _op_data if x not in _deprecated) # type:ignore
_symbol_data = dict((getattr(ast, x, None), y) for x, y, z in _op_data if x not in _deprecated) # type:ignore

class op_util:
    """
    This class provides data and functions for mapping
    AST nodes to symbols and precedences.
    """
    @classmethod
    def get_op_symbol(cls, obj:ast.operator|ast.boolop|ast.cmpop|ast.unaryop,
                      fmt:str='%s', 
                      symbol_data:dict[type[ast.AST]|None, str]=_symbol_data, 
                      type:Callable[[object], type[Any]]=type) -> str:
        """Given an AST node object, returns a string containing the symbol.
        """
        return fmt % symbol_data[type(obj)]
    @classmethod
    def get_op_precedence(cls, obj:ast.AST, 
                          precedence_data:dict[type[ast.AST]|None, int]=_precedence_data, 
                          type:Callable[[object], type[Any]]=type) -> int:
        """Given an AST node object, returns the precedence.

        @raises KeyError: If the node is not explicitely supported by this function. 
            This is a very legacy piece of code, all calls to L{get_op_precedence} should be
            guarded in a C{try:... except KeyError:...} statement.
        """
        return precedence_data[type(obj)]

    if not TYPE_CHECKING:
        class Precedence(object):
            vars().update((cast(str, x), z) for x, _, z in _op_data)
            highest = max(cast(int, z) for _, _, z in _op_data) + 2
    else:
        Precedence: Any

del _op_data, _index, _precedence_data, _symbol_data, _deprecated
# This was part of the astor library for Python AST manipulation.


class _OldSchoolNamespacePackageVis(ast.NodeVisitor):

    is_namespace_package: bool = False

    def visit_Module(self, node: ast.Module) -> None:
        try:
            self.generic_visit(node)
        except StopIteration:
            pass
    
    def visit_skip(self, node: ast.AST) -> None:
        pass
    
    visit_FunctionDef = visit_AsyncFunctionDef = visit_ClassDef = visit_skip
    visit_AugAssign = visit_skip
    visit_Return = visit_Raise = visit_Assert = visit_skip
    visit_Pass = visit_Break = visit_Continue = visit_Delete = visit_skip
    visit_Global = visit_Nonlocal = visit_skip
    visit_Import = visit_ImportFrom = visit_skip

    def visit_Expr(self, node: ast.Expr) -> None:
        # Search for ast.Expr nodes that contains a call to a name or attribute 
        # access of "declare_namespace" and a single argument: __name__
        if not isinstance(val:=node.value, ast.Call):
            return
        if not isinstance(func:=val.func, (ast.Name, ast.Attribute)):
            return
        if isinstance(func, ast.Name) and func.id == 'declare_namespace' or \
           isinstance(func, ast.Attribute) and func.attr == 'declare_namespace':
            # checks the arguments are the basic one, not custom
            try:
                arg1, = (*val.args, *(k.value for k in val.keywords))
            except ValueError:
                raise StopIteration
            if not isinstance(arg1, ast.Name) or arg1.id != '__name__':
                raise StopIteration
            
            self.is_namespace_package = True
            raise StopIteration
        
    def visit_Assign(self, node: ast.Assign) -> None:
        # search for assignments nodes that contains a call in the 
        # rhs to name or attribute acess of "extend_path" and two arguments: 
        # __path__ and __name__. 

        if not any(isinstance(t, ast.Name) and t.id == '__path__' for t in node.targets):
            return
        if not isinstance(val:=node.value, ast.Call):
            return
        if not isinstance(func:=val.func, (ast.Name, ast.Attribute)):
            return
        if isinstance(func, ast.Name) and func.id == 'extend_path' or \
           isinstance(func, ast.Attribute) and func.attr == 'extend_path':
            # checks the arguments are the basic one, not custom
            try:
                arg1, arg2 = (*val.args, *(k.value for k in val.keywords))
            except ValueError:
                raise StopIteration
            if (not isinstance(arg1, ast.Name)) or arg1.id != '__path__':
                raise StopIteration
            if (not isinstance(arg2, ast.Name)) or arg2.id != '__name__':
                raise StopIteration

            self.is_namespace_package = True
            raise StopIteration
    
    def visit_AnnAssign(self, node: ast.AnnAssign) -> None:
        setattr(node, 'targets', [node.target])
        try:
            self.visit_Assign(node) # type:ignore[arg-type]
        finally:
            delattr(node, 'targets')
       
def is_old_school_namespace_package(tree: ast.Module) -> bool:
    """
    Returns True if the module is a pre PEP 420 namespace package::
    
        from pkgutil import extend_path
        __path__ = extend_path(__path__, __name__)
        # OR
        import pkg_resources
        pkg_resources.declare_namespace(__name__)
        # OR
        __import__('pkg_resources').declare_namespace(__name__)
        # OR
        import pkg_resources
        pkg_resources.declare_namespace(name=__name__)

    The following code will return False, tho::

        from pkgutil import extend_path
        __path__ = extend_path(__path__, __name__ + '.impl')
    
    """
    v =_OldSchoolNamespacePackageVis()
    v.visit(tree)
    return v.is_namespace_package



"""
Code building ``all-documents.html``, ``searchindex.json`` and ``fullsearchindex.json``.
"""
from __future__ import annotations

from pathlib import Path
from typing import Iterator, List, Optional, Tuple, Type, Dict, TYPE_CHECKING
import json

import attr

from pydoctor.templatewriter.pages import Page
from pydoctor import model, epydoc2stan

from twisted.web.template import Tag, renderer
from lunr import lunr, get_default_builder

if TYPE_CHECKING:
    from twisted.web.template import Flattenable

def get_all_documents_flattenable(system: model.System) -> Iterator[Dict[str, "Flattenable"]]:
    """
    Get a generator for all data to be writen into ``all-documents.html`` file.
    """
    # This function accounts for a substantial proportion of pydoctor runtime. 
    # So it's optimized.
    insert_break_points = epydoc2stan.insert_break_points
    format_kind = epydoc2stan.format_kind
    format_summary = epydoc2stan.format_summary

    return ({
            'id':         ob.fullName(), 
            'name':       ob.name, 
            'fullName':   insert_break_points(ob.fullName()), 
            'kind':       format_kind(ob.kind) if ob.kind else '', 
            'type':       str(ob.__class__.__name__),
            'summary':    format_summary(ob),
            'url':        ob.url,
            'privacy':    str(ob.privacyClass.name)}   

            for ob in system.allobjects.values() if ob.isVisible)

class AllDocuments(Page):
    
    filename = 'all-documents.html'

    def title(self) -> str:
        return "All Documents"

    @renderer
    def documents(self, request: None, tag: Tag) -> Iterator[Tag]:        
        for doc in get_all_documents_flattenable(self.system):
            yield tag.clone().fillSlots(**doc)

@attr.s(auto_attribs=True)
class LunrIndexWriter:
    """
    Class to write lunr indexes with configurable fields. 
    """
    
    output_file: Path
    system: model.System
    fields: List[str]

    _BOOSTS = {
                'name':6,
                'names': 1,
                'qname':2,
                'docstring':1,
                'kind':-1
              }
    
    # For all pipeline functions, stop_word_filter, stemmer and trimmer, skip their action expect for the
    # docstring field.
    _SKIP_PIPELINES = list(_BOOSTS)
    _SKIP_PIPELINES.remove('docstring')
    
    @staticmethod
    def get_ob_boost(ob: model.Documentable) -> int:
        # Advantage container types because they hold more informations.
        if isinstance(ob, (model.Class, model.Module)):
            return 2
        else:
            return 1
    
    def format(self, ob: model.Documentable, field:str) -> Optional[str]:
        try:
            return getattr(self, f'format_{field}')(ob) #type:ignore[no-any-return]
        except AttributeError as e:
            raise AssertionError() from e
    
    def format_name(self, ob: model.Documentable) -> str:
        return ob.name
    
    def format_names(self, ob: model.Documentable) -> str:
        return ' '.join(stem_identifier(ob.name))
    
    def format_qname(self, ob: model.Documentable) -> str:
        return ob.fullName()
    
    def format_docstring(self, ob: model.Documentable) -> Optional[str]:
        # sanitize docstring in a proper way to be more easily indexable by lunr.
        doc = None
        source = epydoc2stan.ensure_parsed_docstring(ob)
        if source is not None:
            assert ob.parsed_docstring is not None
            doc = ob.parsed_docstring.to_text()
        return doc

    def format_kind(self, ob:model.Documentable) -> str:
        return epydoc2stan.format_kind(ob.kind) if ob.kind else ''

    def get_corpus(self) -> List[Tuple[Dict[str, Optional[str]], Dict[str, int]]]:
        return [
            (
                {
                    f:self.format(ob, f) for f in self.fields
                }, 
                {
                    "boost": self.get_ob_boost(ob)
                }
            )
            for ob in (o for o in self.system.allobjects.values() if o.isVisible)
        ]

    def write(self) -> None:

        builder = get_default_builder()

        # Skip some pipelines for better UX
        # https://lunr.readthedocs.io/en/latest/customisation.html#skip-a-pipeline-function-for-specific-field-names
        
        # We want classes named like "For" to be indexed with their name, even if it's matching stop words.
        # We don't want "name" and related fields to be stemmed since we're stemming ourselves the name.
        # see https://github.com/twisted/pydoctor/issues/648 for why.
        for pipeline_function in builder.pipeline.registered_functions.values():
            builder.pipeline.skip(pipeline_function, self._SKIP_PIPELINES)  

        # Removing the stemmer from the search pipeline, see https://github.com/yeraydiazdiaz/lunr.py/issues/112
        builder.search_pipeline.reset()

        index = lunr(
            ref='qname',
            fields=[{'field_name':name, 'boost':self._BOOSTS[name]} for name in self.fields],
            documents=self.get_corpus(), 
            builder=builder)   
        
        serialized_index = json.dumps(index.serialize())

        with self.output_file.open('w', encoding='utf-8') as fobj:
            fobj.write(serialized_index)

# https://lunr.readthedocs.io/en/latest/
def write_lunr_index(output_dir: Path, system: model.System) -> None:
    """
    Write ``searchindex.json`` and ``fullsearchindex.json`` to the output directory.

    @arg output_dir: Output directory.
    @arg system: System. 
    """
    LunrIndexWriter(output_dir / "searchindex.json", 
        system=system, 
        fields=["name", "names", "qname"]
        ).write()

    LunrIndexWriter(output_dir / "fullsearchindex.json", 
        system=system, 
        fields=["name", "names", "qname", "docstring", "kind"]
        ).write()


def stem_identifier(identifier: str) -> Iterator[str]:
    # we are stemming the identifier ourselves because
    # lunr is removing too much of important data. 
    # See issue https://github.com/twisted/pydoctor/issues/648
    yielded = set()
    parts = epydoc2stan._split_indentifier_parts_on_case(identifier)
    for p in parts:
        p = p.strip('_')
        if p:
            if p not in yielded:
                yielded.add(p)
                yield p

searchpages: List[Type[Page]] = [AllDocuments]



"""The classes that turn  L{Documentable} instances into objects we can render."""
from __future__ import annotations

from typing import (
    TYPE_CHECKING, Dict, Iterator, List, Optional, Mapping, Sequence,
    Type, Union
)
import ast
import abc
from urllib.parse import urljoin

from twisted.web.iweb import IRenderable, ITemplateLoader, IRequest
from twisted.web.template import Element, Tag, renderer, tags, CharRef
from pydoctor.extensions import zopeinterface

from pydoctor import epydoc2stan, model, __version__
from pydoctor.astbuilder import node2fullname
from pydoctor.templatewriter import util, TemplateLookup, TemplateElement
from pydoctor.templatewriter.pages.table import ChildTable
from pydoctor.templatewriter.pages.sidebar import SideBar
from pydoctor.epydoc.markup._pyval_repr import colorize_inline_pyval

if TYPE_CHECKING:
    from typing_extensions import Final
    from twisted.web.template import Flattenable
    from pydoctor.templatewriter.pages.attributechild import AttributeChild
    from pydoctor.templatewriter.pages.functionchild import FunctionChild


def _format_decorators(obj: Union[model.Function, model.Attribute, model.FunctionOverload]) -> Iterator["Flattenable"]:
    # Since we use this function to colorize the FunctionOverload decorators and it's not an actual Documentable subclass, we use the overload's 
    # primary function for parts that requires an interface to Documentable methods or attributes
    documentable_obj = obj if not isinstance(obj, model.FunctionOverload) else obj.primary

    for dec in obj.decorators or ():
        if isinstance(dec, ast.Call):
            fn = node2fullname(dec.func, documentable_obj)
            # We don't want to show the deprecated decorator;
            # it shows up as an infobox.
            # TODO: move this somewhere it can be customized.
            if fn in ("twisted.python.deprecate.deprecated",
                      "twisted.python.deprecate.deprecatedProperty"):
                break

        # Colorize decorators!
        doc = colorize_inline_pyval(dec)
        stan = epydoc2stan.safe_to_stan(doc, documentable_obj.docstring_linker, documentable_obj,
            fallback=epydoc2stan.colorized_pyval_fallback, 
            section='rendering of decorators')
        
        # Report eventual warnings. It warns when we can't colorize the expression for some reason.
        epydoc2stan.reportWarnings(documentable_obj, doc.warnings, section='colorize decorator')
        
        yield tags.span('@', stan.children, tags.br(), class_='decorator')

def format_decorators(obj: Union[model.Function, model.Attribute, model.FunctionOverload]) -> Tag:
    if decs:=list(_format_decorators(obj)):
        return tags.div(decs)
    return tags.transparent

def format_signature(func: Union[model.Function, model.FunctionOverload]) -> "Flattenable":
    """
    Return a stan representation of a nicely-formatted source-like function signature for the given L{Function}.
    Arguments default values are linked to the appropriate objects when possible.
    """

    parsed_sig = epydoc2stan.get_parsed_signature(func)
    if parsed_sig is None:
        return "(...)"
    ctx = func.primary if isinstance(func, model.FunctionOverload) else func
    return epydoc2stan.safe_to_stan(
        parsed_sig, 
        ctx.docstring_linker, 
        ctx, 
        fallback=lambda _, doc, ___: tags.transparent(doc.to_text()),
        section='signature'
    )

def format_class_signature(cls: model.Class) -> "Flattenable":
    """
    The class signature is the formatted list of bases this class extends. 
    It's not the class constructor.
    """
    r: List["Flattenable"] = []
    # the linker will only be used to resolve the generic arguments of the base classes, 
    # it won't actually resolve the base classes (see comment few lines below).
    # this is why we're using the annotation linker.
    _linker = cls.docstring_linker
    if cls.rawbases:
        r.append('(')
        
        for idx, ((str_base, base_node), base_obj) in enumerate(zip(cls.rawbases, cls.baseobjects)):
            if idx != 0:
                r.append(', ')

            # Make sure we bypass the linker’s resolver process for base object, 
            # because it has been resolved already (with two passes).
            # Otherwise, since the class declaration wins over the imported names,
            # a class with the same name as a base class confused pydoctor and it would link 
            # to it self: https://github.com/twisted/pydoctor/issues/662

            refmap = None
            if base_obj is not None:
                refmap = {str_base:base_obj.fullName()}
                
            # link to external class, using the colorizer here
            # to link to classes with generics (subscripts and other AST expr).
            # we use is_annotation=True because bases are unstringed, they can contain annotations. 
            stan = epydoc2stan.safe_to_stan(colorize_inline_pyval(base_node, refmap=refmap, is_annotation=True), _linker, cls, 
                fallback=epydoc2stan.colorized_pyval_fallback, 
                section='rendering of class signature')
            r.extend(stan.children)
                
        r.append(')')
    return r

LONG_SIGNATURE = 88 # this doesn't acount for the 'def ' and the ending ':'
"""
Maximum size of a function definition to be rendered on a single line. 
The multiline formatting is only applied at the CSS level to stay customizable. 
We add a css class to the signature HTML to signify the signature could possibly
be better formatted on several lines.
"""
def format_overloads(func: model.Function) -> Iterator["Flattenable"]:
    """
    Format a function overloads definitions as nice HTML signatures.
    """

    for overload in func.overloads:
        yield tags.div(format_decorators(overload), 
            tags.div(format_function_def(func.name, func.is_async, overload)),   
            class_='function-overload')

_nbsp = CharRef(160) # non-breaking space.
def format_function_def(func_name: str, is_async: bool, 
                        func: Union[model.Function, model.FunctionOverload]) -> List["Flattenable"]:
    """
    Format a function definition as nice HTML signature. 
    
    If the function is overloaded, it will return an empty list. 
    We use L{format_overloads} for these.
    """
    r:List["Flattenable"] = []
    # If this is a function with overloads, we do not render the principal 
    # signature because the overloaded signatures will be shown instead.
    if isinstance(func, model.Function) and func.overloads:
        return r
    def_stmt = ['async', _nbsp, 'def'] if is_async else ['def']
    if func_name.endswith('.setter') or func_name.endswith('.deleter'):
        func_name = func_name[:func_name.rindex('.')]
    
    func_signature_css_class = 'function-signature'
    
    # We never mark the overloaded functions as long since this could make the output of pydoctor
    # worst that before when there are many overloads to be wrapped. It allows to
    # to scroll less to get to the actual main documentation of the function.
    if not isinstance(func, model.FunctionOverload) and \
        epydoc2stan.function_signature_len(func) > LONG_SIGNATURE:
        func_signature_css_class += ' long-signature'
    
    r.extend([
        tags.span(def_stmt, class_='py-keyword'), _nbsp,
        tags.span(func_name, class_='py-defname'), 
        tags.span(format_signature(func), ':', 
                  class_=func_signature_css_class),
    ])
    return r
    

class Nav(TemplateElement):
    """
    Common navigation header.
    """

    filename = 'nav.html'

class Head(TemplateElement):
    """
    Common metadata.
    """

    filename = 'head.html'

    def __init__(self, title: str, baseurl: str | None, pageurl: str, 
                 loader: ITemplateLoader) -> None:
        super().__init__(loader)
        self._title = title
        self._baseurl = baseurl
        self._pageurl = pageurl
    
    @renderer
    def canonicalurl(self, request: IRequest, tag: Tag) -> Flattenable:
        if not self._baseurl:
            return ''
        canonical_link = urljoin(self._baseurl, self._pageurl)
        return tags.link(rel='canonical', href=canonical_link)

    @renderer
    def title(self, request: IRequest, tag: Tag) -> str:
        return self._title


class Page(TemplateElement):
    """
    Abstract base class for output pages.

    Defines special HTML placeholders that are designed to be overriden by users:
    "header.html", "subheader.html" and "footer.html".
    """

    def __init__(self, system: model.System,
                 template_lookup: TemplateLookup,
                 loader: Optional[ITemplateLoader] = None):
        self.system = system
        self.template_lookup = template_lookup
        if not loader:
            loader = self.lookup_loader(template_lookup)
        super().__init__(loader)
    
    @property
    def page_url(self) -> str:
        # This MUST be overriden in CommonPage
        """
        The relative page url
        """
        return self.filename

    def render(self, request: Optional[IRequest]) -> Tag:
        return tags.transparent(super().render(request)).fillSlots(**self.slot_map)

    @property
    def slot_map(self) -> Dict[str, "Flattenable"]:
        system = self.system

        if system.options.projecturl:
            project_tag = tags.a(href=system.options.projecturl, class_="projecthome")
        else:
            project_tag = tags.transparent
        project_tag(system.projectname)

        return dict(
            project=project_tag,
            pydoctor_version=__version__,
            buildtime=system.buildtime.strftime("%Y-%m-%d %H:%M:%S"),
        )

    @abc.abstractmethod
    def title(self) -> str:
        raise NotImplementedError()

    @renderer
    def head(self, request: IRequest, tag: Tag) -> IRenderable:
        return Head(self.title(), self.system.options.htmlbaseurl, self.page_url, 
                    loader=Head.lookup_loader(self.template_lookup))

    @renderer
    def nav(self, request: IRequest, tag: Tag) -> IRenderable:
        return Nav(Nav.lookup_loader(self.template_lookup))

    @renderer
    def header(self, request: IRequest, tag: Tag) -> IRenderable:
        return Element(self.template_lookup.get_loader('header.html'))

    @renderer
    def subheader(self, request: IRequest, tag: Tag) -> IRenderable:
        return Element(self.template_lookup.get_loader('subheader.html'))

    @renderer
    def footer(self, request: IRequest, tag: Tag) -> IRenderable:
        return Element(self.template_lookup.get_loader('footer.html'))


class CommonPage(Page):

    filename = 'common.html'
    ob: model.Documentable

    def __init__(self, ob: model.Documentable, template_lookup: TemplateLookup, docgetter: Optional[util.DocGetter]=None):
        super().__init__(ob.system, template_lookup)
        self.ob = ob
        if docgetter is None:
            docgetter = util.DocGetter()
        self.docgetter = docgetter
        self._order = ob.system.membersOrder(ob)

    @property
    def page_url(self) -> str:
        return self.ob.page_object.url

    def title(self) -> str:
        return self.ob.fullName()

    def heading(self) -> Tag:
        return tags.h1(class_=util.css_class(self.ob))(
            tags.code(self.namespace(self.ob))
            )

    def category(self) -> str:
        kind = self.ob.kind
        assert kind is not None
        return f"{epydoc2stan.format_kind(kind).lower()} documentation"

    def namespace(self, obj: model.Documentable) -> List[Union[Tag, str]]:
        page_url = self.page_url
        parts: List[Union[Tag, str]] = []
        ob: Optional[model.Documentable] = obj
        while ob:
            if ob.documentation_location is model.DocLocation.OWN_PAGE:
                if parts:
                    parts.extend(['.', tags.wbr])
                parts.append(tags.code(epydoc2stan.taglink(ob, page_url, ob.name)))
            ob = ob.parent
        parts.reverse()
        return parts
    
    @renderer
    def inhierarchy(self, request: object, tag: Tag) -> "Flattenable":
        return ()

    def extras(self) -> List["Flattenable"]:
        return self.objectExtras(self.ob)

    def docstring(self) -> "Flattenable":
        return self.docgetter.get(self.ob)

    def children(self) -> Sequence[model.Documentable]:
        return sorted(
            (o for o in self.ob.contents.values() if o.isVisible),
            key=self._order)

    def packageInitTable(self) -> "Flattenable":
        return ()

    @renderer
    def baseTables(self, request: object, tag: Tag) -> "Flattenable":
        return ()

    def mainTable(self) -> "Flattenable":
        children = self.children()
        if children:
            return ChildTable(self.docgetter, self.ob, children,
                    ChildTable.lookup_loader(self.template_lookup))
        else:
            return ()

    def methods(self) -> Sequence[model.Documentable]:
        return sorted((o for o in self.ob.contents.values()
                       if o.documentation_location is model.DocLocation.PARENT_PAGE and o.isVisible), 
                      key=self._order)

    def childlist(self) -> List[Union["AttributeChild", "FunctionChild"]]:
        from pydoctor.templatewriter.pages.attributechild import AttributeChild
        from pydoctor.templatewriter.pages.functionchild import FunctionChild

        r: List[Union["AttributeChild", "FunctionChild"]] = []

        func_loader = FunctionChild.lookup_loader(self.template_lookup)
        attr_loader = AttributeChild.lookup_loader(self.template_lookup)

        for c in self.methods():
            if isinstance(c, model.Function):
                r.append(FunctionChild(self.docgetter, c, self.objectExtras(c), func_loader))
            elif isinstance(c, model.Attribute):
                r.append(AttributeChild(self.docgetter, c, self.objectExtras(c), attr_loader))
            else:
                assert False, type(c)
        return r

    def objectExtras(self, ob: model.Documentable) -> List["Flattenable"]:
        """
        Flatten each L{model.Documentable.extra_info} list item.
        """
        r: List["Flattenable"] = []
        for extra in ob.extra_info:
            r.append(epydoc2stan.unwrap_docstring_stan(
                epydoc2stan.safe_to_stan(extra, ob.docstring_linker, ob,
                fallback = lambda _,__,___:epydoc2stan.BROKEN, section='extra')))
        return r


    def functionBody(self, ob: model.Documentable) -> "Flattenable":
        return self.docgetter.get(ob)

    @renderer
    def maindivclass(self, request: IRequest, tag: Tag) -> str:
        return 'nosidebar' if self.ob.system.options.nosidebar else ''

    @renderer
    def sidebarcontainer(self, request: IRequest, tag: Tag) -> Union[Tag, str]:
        if self.ob.system.options.nosidebar:
            return ""
        else:
            return tag.fillSlots(sidebar=SideBar(ob=self.ob, template_lookup=self.template_lookup))

    @property
    def slot_map(self) -> Dict[str, "Flattenable"]:
        slot_map = super().slot_map
        slot_map.update(
            heading=self.heading(),
            category=self.category(),
            extras=self.extras(),
            docstring=self.docstring(),
            mainTable=self.mainTable(),
            packageInitTable=self.packageInitTable(),
            childlist=self.childlist(),
        )
        return slot_map

def source_tag(href: str) -> Tag: 
    return tags.a("(source)", href=href, class_="sourceLink")

class ModulePage(CommonPage):
    ob: model.Module

    def source_links(self) -> Flattenable | None:
        if sourceHref:=util.srclink(self.ob):
            return source_tag(sourceHref)
        return None

    def extras(self) -> List["Flattenable"]:
        r: List["Flattenable"] = []
        if links:=self.source_links():
            r.append(links)
        r.extend(super().extras())
        return r


class PackagePage(ModulePage):
    ob: model.Package

    def children(self) -> Sequence[model.Documentable]:
        return sorted(self.ob.submodules(), key=self._order)

    def packageInitTable(self) -> "Flattenable":
        children = sorted(
            (o for o in self.ob.contents.values()
             if not isinstance(o, model.Module) and o.isVisible),
            key=self._order)
        if children:
            loader = ChildTable.lookup_loader(self.template_lookup)
            return [
                tags.p("From ", tags.code("__init__.py"), ":", class_="fromInitPy"),
                ChildTable(self.docgetter, self.ob, children, loader)
                ]
        else:
            return ()
    
    def source_links(self) -> Flattenable | None:
        # supports multiple source links, since there could be multiple source paths
        # for namespace packages
        links = util.package_srclinks(self.ob)
        links_max_index = len(links) - 1
        if links_max_index == -1:
            return None
        r: list[Flattenable] = []
        for i, href in enumerate(links):
            r.append(source_tag(href))
            if 0 <= i < links_max_index:
                r.append(', ')
        return tags.transparent(*r)

    def methods(self) -> Sequence[model.Documentable]:
        return sorted([o for o in self.ob.contents.values()
                if o.documentation_location is model.DocLocation.PARENT_PAGE
                and o.isVisible], key=self._order)

def assembleList(
        system: model.System,
        label: str,
        lst: Sequence[str],
        page_url: str
        ) -> Optional["Flattenable"]:
    """
    Convert list of object names into a stan tree with clickable links. 
    """
    lst2 = []
    for name in lst:
        o = system.allobjects.get(name)
        if o is None or o.isVisible:
            lst2.append(name)
    lst = lst2
    if not lst:
        return None
    def one(item: str) -> "Flattenable":
        if item in system.allobjects:
            return tags.code(epydoc2stan.taglink(system.allobjects[item], page_url))
        else:
            return item
    def commasep(items: Sequence[str]) -> List["Flattenable"]:
        r = []
        for item in items:
            r.append(one(item))
            r.append(', ')
        del r[-1]
        return r
    p: List["Flattenable"] = [label]
    p.extend(commasep(lst))
    return p


class ClassPage(CommonPage):

    ob: model.Class

    def __init__(self,
            ob: model.Documentable,
            template_lookup: TemplateLookup,
            docgetter: Optional[util.DocGetter] = None
            ):
        super().__init__(ob, template_lookup, docgetter)
        self.baselists = util.class_members(self.ob)

    def extras(self) -> List["Flattenable"]:
        r: List["Flattenable"] = []

        sourceHref = util.srclink(self.ob)
        source: "Flattenable"
        if sourceHref:
            source = (" ", tags.a("(source)", href=sourceHref, class_="sourceLink"))
        else:
            source = tags.transparent
        r.append(tags.p(tags.code(
            tags.span("class", class_='py-keyword'), " ",
            tags.span(self.ob.name, class_='py-defname'),
            self.classSignature(), ":", source
            ), class_='class-signature'))

        subclasses = sorted(self.ob.subclasses, key=util.alphabetical_order_func)
        if subclasses:
            p = assembleList(self.ob.system, "Known subclasses: ",
                            [o.fullName() for o in subclasses], self.page_url)
            if p is not None:
                r.append(tags.p(p))

        constructor = epydoc2stan.get_constructors_extra(self.ob)
        if constructor:
            r.append(epydoc2stan.unwrap_docstring_stan(
                epydoc2stan.safe_to_stan(constructor, self.ob.docstring_linker, self.ob,
                fallback = lambda _,__,___:epydoc2stan.BROKEN, section='constructor extra')))

        r.extend(super().extras())
        return r

    def classSignature(self) -> "Flattenable":
        return format_class_signature(self.ob)

    @renderer
    def inhierarchy(self, request: object, tag: Tag) -> Tag:
        return tag(href="classIndex.html#"+self.ob.fullName())

    @renderer
    def baseTables(self, request: object, item: Tag) -> "Flattenable":
        baselists = self.baselists[:]
        if not baselists:
            return []
        if baselists[0][0][0] == self.ob:
            del baselists[0]
        loader = ChildTable.lookup_loader(self.template_lookup)
        return [item.clone().fillSlots(
                          baseName=self.baseName(b),
                          baseTable=ChildTable(self.docgetter, self.ob,
                                               sorted(attrs, key=self._order),
                                               loader))
                for b, attrs in baselists]

    def baseName(self, bases: Sequence[model.Class]) -> "Flattenable":
        page_url = self.page_url
        r: List["Flattenable"] = []
        source_base = bases[0]
        r.append(tags.code(epydoc2stan.taglink(source_base, page_url, source_base.name)))
        bases_to_mention = bases[1:-1]
        if bases_to_mention:
            tail: List["Flattenable"] = []
            for b in reversed(bases_to_mention):
                tail.append(tags.code(epydoc2stan.taglink(b, page_url, b.name)))
                tail.append(', ')
            del tail[-1]
            r.extend([' (via ', tail, ')'])
        return r

    def objectExtras(self, ob: model.Documentable) -> List["Flattenable"]:
        r: List["Flattenable"] = list(get_override_info(self.ob, ob.name, self.page_url))
        r.extend(super().objectExtras(ob))
        return r

def get_override_info(cls:model.Class, member_name:str, page_url:Optional[str]=None) -> Iterator["Flattenable"]:
    page_url = page_url or cls.page_object.url
    for b in cls.mro(include_self=False):
        if member_name not in b.contents:
            continue
        overridden = b.contents[member_name]
        yield tags.div(class_="interfaceinfo")(
            'overrides ', tags.code(epydoc2stan.taglink(overridden, page_url)))
        break
    
    ocs = sorted(util.overriding_subclasses(cls, member_name), key=util.alphabetical_order_func)
    if ocs:
        l = assembleList(cls.system, 'overridden in ',
                            [o.fullName() for o in ocs], page_url)
        if l is not None:
            yield tags.div(class_="interfaceinfo")(l)
    

class ZopeInterfaceClassPage(ClassPage):
    ob: zopeinterface.ZopeInterfaceClass

    def extras(self) -> List["Flattenable"]:
        r = super().extras()
        if self.ob.isinterface:
            namelist = [o.fullName() for o in 
                        sorted(self.ob.implementedby_directly, key=util.alphabetical_order_func)]
            label = 'Known implementations: '
        else:
            namelist = sorted(self.ob.implements_directly, key=lambda x:x.lower())
            label = 'Implements interfaces: '
        if namelist:
            l = assembleList(self.ob.system, label, namelist, self.page_url)
            if l is not None:
                r.append(tags.p(l))
        return r

    def interfaceMeth(self, methname: str) -> Optional[model.Documentable]:
        system = self.ob.system
        for interface in self.ob.allImplementedInterfaces:
            if interface in system.allobjects:
                io = system.allobjects[interface]
                assert isinstance(io, zopeinterface.ZopeInterfaceClass)
                for io2 in io.mro():
                    method: Optional[model.Documentable] = io2.contents.get(methname)
                    if method is not None:
                        return method
        return None

    def objectExtras(self, ob: model.Documentable) -> List["Flattenable"]:
        imeth = self.interfaceMeth(ob.name)
        r: List["Flattenable"] = []
        if imeth:
            iface = imeth.parent
            assert iface is not None
            r.append(tags.div(class_="interfaceinfo")('from ', tags.code(
                epydoc2stan.taglink(imeth, self.page_url, iface.fullName())
                )))
        r.extend(super().objectExtras(ob))
        return r

commonpages: 'Final[Mapping[str, Type[CommonPage]]]' = {
    'Module': ModulePage,
    'Package': PackagePage,
    'Class': ClassPage,
    'ZopeInterfaceClass': ZopeInterfaceClassPage,
}
"""List all page classes: ties documentable class name with the page class used for rendering"""
